/*
 *  This file is part of x48, an emulator of the HP-48sx Calculator.
 *  Copyright (C) 1994  Eddie C. Dost  (ecd@dressler.de)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Revision 1.1  1994/08/26  11:09:02  ecd
 * Initial revision
 *
 * $Id: emulate.c,v 1.5 1994/09/18 22:47:20 ecd Exp ecd $
 */


#include "global.h"

#include <stdio.h>
#include <sys/time.h>

#include "hp48.h"
#include "hp48_emu.h"
#include "device.h"
#include "debug.h"
#include "x48_x11.h"

#ifdef SOLARIS
extern int gettimeofday __ProtoType__((struct timeval *tp));
#endif

static long jumpaddr;

#ifdef DEBUG_PRINT
static char *std_fields[] = {
  "P", "WP", "XS", "X", "S", "M", "B", "W",
  "P", "WP", "XS", "X", "S", "M", "B", "A"
};
#endif

#ifdef SINGLE_STEP
void
#ifdef __FunctionProto__
wait_key(void)
#else
wait_key()
#endif
{
  char buf[10];

  while (fread(buf, 1, 1, stdin) != 1) ;
}
#endif

int
#ifdef __FunctionProto__
decode_group_80(void)
#else
decode_group_80()
#endif
{
  long t, op3, op4, op5, op6;
  unsigned char *REG;
  long addr;

  op3 = read_nibble(saturn.PC + 2);
  switch (op3) {
    case 0:	/* OUT=CS */
#ifdef DEBUG_PRINT
      dprintf("%.5lx\tOUT=CS\n", saturn.PC);
#endif
      saturn.PC += 3;
      copy_register(saturn.OUT, saturn.C, OUTS_FIELD);
#ifdef DUMP_REGISTERS
      dump_register(saturn.OUT);
#endif
      return 0;
    case 1:	/* OUT=C */
#ifdef DEBUG_PRINT
      dprintf("%.5lx\tOUT=C\n", saturn.PC);
#endif
      saturn.PC += 3;
      copy_register(saturn.OUT, saturn.C, OUT_FIELD);
#ifdef DUMP_REGISTERS
      dump_register(saturn.OUT);
#endif
      return 0;
    case 2:	/* A=IN */
#ifdef DEBUG_PRINT
      dprintf("%.5lx\tA=IN\n", saturn.PC);
#endif
      saturn.PC += 3;
      do_in();
      copy_register(saturn.A, saturn.IN, IN_FIELD);
#ifdef DUMP_REGISTERS
      dump_register(saturn.A);
#endif
      return 0;
    case 3:	/* C=IN */
#ifdef DEBUG_PRINT
      dprintf("%.5lx\tC=IN\n", saturn.PC);
#endif
      saturn.PC += 3;
      do_in();
      copy_register(saturn.C, saturn.IN, IN_FIELD);
#ifdef DUMP_REGISTERS
      dump_register(saturn.C);
#endif
      return 0;
    case 4:	/* UNCNFG */
#ifdef DEBUG_PRINT
      dprintf("%.5lx\tUNCNFG\n", saturn.PC);
#endif
      saturn.PC += 3;
      do_unconfigure();
      return 0;
    case 5:	/* CONFIG */
#ifdef DEBUG_PRINT
      dprintf("%.5lx\tCONFIG\n", saturn.PC);
#endif
      saturn.PC += 3;
      do_configure();
      return 0;
    case 6:	/* C=ID */
#ifdef DEBUG_PRINT
      dprintf("%.5lx\tC=ID\n", saturn.PC);
#endif
      saturn.PC += 3;
      return get_identification();
    case 7:	/* SHUTDN */
#ifdef DEBUG_PRINT
      dprintf("%.5lx\tSHUTDN\n", saturn.PC);
#endif
      saturn.PC += 3;
      do_shutdown();
      return 0;
    case 8:
      op4 = read_nibble(saturn.PC + 3);
      switch (op4) {
        case 0:		/* INTON */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tINTON\n", saturn.PC);
#endif
          saturn.PC += 4;
          do_inton();
#ifdef DEBUG_INTERUPT
	  fprintf(stderr, "INTERUPT ON\n");
#endif
	  return 0;
        case 1:		/* RSI... */
          op5 = read_nibble(saturn.PC + 4);
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tRSI%s\n", saturn.PC, op5 ? "*" : "");
#endif
          saturn.PC += 5;
          do_reset_interrupt_system();
#ifdef DEBUG_INTERUPT
          fprintf(stderr, "saturn.PC = %.5lx: RSI\n", saturn.PC);
#endif
	  return 0;
        case 2:		/* LA... */
          op5 = read_nibble(saturn.PC + 4);
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tLAHEX\t%s\n", saturn.PC,
                  make_hexstr(saturn.PC + 5, op5 + 1));
#endif
          load_constant(saturn.A, op5 + 1, saturn.PC + 5);
          saturn.PC += 6 + op5;
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
#endif
	  return 0;
        case 3:		/* BUSCB */
          saturn.PC += 4;
	  return 0;
        case 4:		/* ABIT=0 */
          op5 = read_nibble(saturn.PC + 4);
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tABIT=0\t%d\n", saturn.PC, op5);
#endif
          saturn.PC += 5;
          clear_register_bit(saturn.A, op5);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.A);
#endif
	  return 0;
        case 5:		/* ABIT=1 */
          op5 = read_nibble(saturn.PC + 4);
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tABIT=1\t%d\n", saturn.PC, op5);
#endif
          saturn.PC += 5;
          set_register_bit(saturn.A, op5);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.A);
#endif
	  return 0;
        case 8:		/* CBIT=0 */
          op5 = read_nibble(saturn.PC + 4);
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tCBIT=0\t%d\n", saturn.PC, op5);
#endif
          saturn.PC += 5;
          clear_register_bit(saturn.C, op5);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.C);
#endif
	  return 0;
        case 9:		/* CBIT=1 */
          op5 = read_nibble(saturn.PC + 4);
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tCBIT=1\t%d\n", saturn.PC, op5);
#endif
          saturn.PC += 5;
          set_register_bit(saturn.C, op5);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.C);
#endif
	  return 0;
        case 6:		/* ?ABIT=0 */
        case 7:		/* ?ABIT=1 */
        case 0xa:	/* ?CBIT=0 */
        case 0xb:	/* ?CBIT=1 */
          op5 = read_nibble(saturn.PC + 4);
	  if (op4 < 8)
            REG = saturn.A;
          else
            REG = saturn.C;
          if (op4 == 6 || op4 == 0xa) 
            t = 0;
          else
            t = 1;
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\t?%cBIT=%d\t%d\n", saturn.PC,
                          (op4<8)?'A':'C', t, op5);
#endif
#ifdef DUMP_REGISTERS
	  dump_register(REG);
#endif
          saturn.CARRY = (get_register_bit(REG, op5) == t)?1:0;
	  if (saturn.CARRY) {
	    saturn.PC += 5;
            op6 = read_nibbles(saturn.PC, 2);
      	    if (op6) {
              if (op6 & 0x80)
                op6 |= 0xffffff00;
              jumpaddr = (saturn.PC + op6) & 0xfffff;
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tGOYES\t%.5lx\n", saturn.PC, jumpaddr);
#endif
	      saturn.PC = jumpaddr;
            } else {
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tRTNYES\n", saturn.PC);
#endif
	      saturn.PC = pop_return_addr();
            }
	  } else {
            saturn.PC += 7;
	  }
          return 0;
        case 0xc:	/* PC=(A) */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tPC=(A)\n", saturn.PC);
#endif
          addr = dat_to_addr(saturn.A);
          jumpaddr = read_nibbles(addr, 5);
          saturn.PC = jumpaddr;
#ifdef DEBUG_PRINT
          dprintf("saturn.PC = %.5lx\n", saturn.PC);
#endif
          return 0;
        case 0xd:	/* BUSCD */
          saturn.PC += 4;
	  return 0;
        case 0xe:	/* PC=(C) */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tPC=(C)\n", saturn.PC);
#endif
          addr = dat_to_addr(saturn.C);
          jumpaddr = read_nibbles(addr, 5);
          saturn.PC = jumpaddr;
#ifdef DEBUG_PRINT
          dprintf("saturn.PC = %.5lx\n", saturn.PC);
#endif
          return 0;
        case 0xf:	/* INTOFF */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tINTOFF\n", saturn.PC);
#endif
          saturn.PC += 4;
          do_intoff();
#ifdef DEBUG_INTERUPT
	  fprintf(stderr, "INTERUPT OFF\n");
#endif
	  return 0;
        default:
          return 1;
      }
    case 9:	/* C+P+1 */
#ifdef DEBUG_PRINT
      dprintf("%.5lx\tC+P+1\n", saturn.PC);
#endif
      saturn.PC += 3;
      add_p_plus_one(saturn.C);
      return 0;
    case 0xa:	/* RESET */
#ifdef DEBUG_PRINT
      dprintf("%.5lx\tRESET\n", saturn.PC);
#endif
      saturn.PC += 3;
      do_reset();
      return 0;
    case 0xb:	/* BUSCC */
      saturn.PC += 3;
      return 0;
    case 0xc:   /* C=P n */
      op4 = read_nibble(saturn.PC + 3);
#ifdef DEBUG_PRINT
      dprintf("%.5lx\tC=P\t%d\n", saturn.PC, op4);
#endif
      saturn.PC += 4;
      set_register_nibble(saturn.C, op4, saturn.P);
      return 0;
    case 0xd:	/* P=C n */
      op4 = read_nibble(saturn.PC + 3);
#ifdef DEBUG_PRINT
      dprintf("%.5lx\tP=C\t%d\n", saturn.PC, op4);
#endif
      saturn.PC += 4;
      saturn.P = get_register_nibble(saturn.C, op4);
      return 0;
    case 0xe:	/* SREQ? */
      return 1;
    case 0xf:	/* CPEX n */
      op4 = read_nibble(saturn.PC + 3);
#ifdef DEBUG_PRINT
      dprintf("%.5lx\tCPEX\t%d\n", saturn.PC, op4);
#endif
      saturn.PC += 4;
      t = get_register_nibble(saturn.C, op4);
      set_register_nibble(saturn.C, op4, saturn.P);
      saturn.P = t;
      return 0;
    default:
      return 1;
  }
}

int
#ifdef __FunctionProto__
decode_group_1(void)
#else
decode_group_1()
#endif
{
  long op, op2, op3, op4;

  op2 = read_nibble(saturn.PC + 1);
  switch (op2) {
    case 0:
      op3 = read_nibble(saturn.PC + 2);
      switch (op3) {
        case 0:		/* saturn.R0=A */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tsaturn.R0=A\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.R0, saturn.A, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.R0);
#endif
          return 0;
        case 1:		/* saturn.R1=A */
        case 5:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tsaturn.R1=A\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.R1, saturn.A, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.R1);
#endif
          return 0;
        case 2:		/* saturn.R2=A */
        case 6:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tsaturn.R2=A\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.R2, saturn.A, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.R2);
#endif
          return 0;
        case 3:		/* saturn.R3=A */
        case 7:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tsaturn.R3=A\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.R3, saturn.A, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.R3);
#endif
          return 0;
        case 4:		/* saturn.R4=A */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tsaturn.R4=A\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.R4, saturn.A, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.R4);
#endif
          return 0;
        case 8:		/* saturn.R0=C */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tsaturn.R0=C\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.R0, saturn.C, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.R0);
#endif
          return 0;
        case 9:		/* saturn.R1=C */
        case 0xd:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tsaturn.R1=C\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.R1, saturn.C, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.R1);
#endif
          return 0;
        case 0xa:	/* saturn.R2=C */
        case 0xe:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tsaturn.R2=C\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.R2, saturn.C, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.R2);
#endif
          return 0;
        case 0xb:	/* saturn.R3=C */
        case 0xf:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tsaturn.R3=C\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.R3, saturn.C, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.R3);
#endif
          return 0;
        case 0xc:	/* saturn.R4=C */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tsaturn.R4=C\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.R4, saturn.C, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.R4);
#endif
          return 0;
        default:
          return 1;
      }
    case 1:
      op3 = read_nibble(saturn.PC + 2);
      switch (op3) {
        case 0:		/* A=saturn.R0 */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tA=saturn.R0\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.A, saturn.R0, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
#endif
          return 0;
        case 1:		/* A=saturn.R1 */
        case 5:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tA=saturn.R1\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.A, saturn.R1, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
#endif
          return 0;
        case 2:		/* A=saturn.R2 */
        case 6:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tA=saturn.R2\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.A, saturn.R2, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
#endif
          return 0;
        case 3:		/* A=saturn.R3 */
        case 7:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tA=saturn.R3\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.A, saturn.R3, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
#endif
          return 0;
        case 4:		/* A=saturn.R4 */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tA=saturn.R4\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.A, saturn.R4, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
#endif
          return 0;
        case 8:		/* C=saturn.R0 */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tC=saturn.R0\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.C, saturn.R0, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
#endif
          return 0;
        case 9:		/* C=saturn.R1 */
        case 0xd:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tC=saturn.R1\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.C, saturn.R1, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
#endif
          return 0;
        case 0xa:	/* C=saturn.R2 */
        case 0xe:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tC=saturn.R2\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.C, saturn.R2, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
#endif
          return 0;
        case 0xb:	/* C=saturn.R3 */
        case 0xf:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tC=saturn.R3\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.C, saturn.R3, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
#endif
          return 0;
        case 0xc:	/* C=saturn.R4 */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tC=saturn.R4\n", saturn.PC);
#endif
          saturn.PC += 3;
          copy_register(saturn.C, saturn.R4, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
#endif
          return 0;
        default:
          return 1;
      }
    case 2:
      op3 = read_nibble(saturn.PC + 2);
      switch (op3) {
        case 0:		/* Asaturn.R0EX */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tAsaturn.R0EX\n", saturn.PC);
#endif
          saturn.PC += 3;
          exchange_register(saturn.A, saturn.R0, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
          dump_register(saturn.R0);
#endif
          return 0;
        case 1:		/* Asaturn.R1EX */
        case 5:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tAsaturn.R1EX\n", saturn.PC);
#endif
          saturn.PC += 3;
          exchange_register(saturn.A, saturn.R1, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
          dump_register(saturn.R1);
#endif
          return 0;
        case 2:		/* Asaturn.R2EX */
        case 6:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tAsaturn.R2EX\n", saturn.PC);
#endif
          saturn.PC += 3;
          exchange_register(saturn.A, saturn.R2, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
          dump_register(saturn.R2);
#endif
          return 0;
        case 3:		/* Asaturn.R3EX */
        case 7:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tAsaturn.R3EX\n", saturn.PC);
#endif
          saturn.PC += 3;
          exchange_register(saturn.A, saturn.R3, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
          dump_register(saturn.R3);
#endif
          return 0;
        case 4:		/* Asaturn.R4EX */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tAsaturn.R4EX\n", saturn.PC);
#endif
          saturn.PC += 3;
          exchange_register(saturn.A, saturn.R4, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
          dump_register(saturn.R4);
#endif
          return 0;
        case 8:		/* Csaturn.R0EX */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tCsaturn.R0EX\n", saturn.PC);
#endif
          saturn.PC += 3;
          exchange_register(saturn.C, saturn.R0, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
          dump_register(saturn.R0);
#endif
          return 0;
        case 9:		/* Csaturn.R1EX */
        case 0xd:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tCsaturn.R1EX\n", saturn.PC);
#endif
          saturn.PC += 3;
          exchange_register(saturn.C, saturn.R1, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
          dump_register(saturn.R1);
#endif
          return 0;
        case 0xa:	/* Csaturn.R2EX */
        case 0xe:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tCsaturn.R2EX\n", saturn.PC);
#endif
          saturn.PC += 3;
          exchange_register(saturn.C, saturn.R2, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
          dump_register(saturn.R2);
#endif
          return 0;
        case 0xb:	/* Csaturn.R3EX */
        case 0xf:
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tCsaturn.R3EX\n", saturn.PC);
#endif
          saturn.PC += 3;
          exchange_register(saturn.C, saturn.R3, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
          dump_register(saturn.R3);
#endif
          return 0;
        case 0xc:	/* Csaturn.R4EX */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tCsaturn.R4EX\n", saturn.PC);
#endif
          saturn.PC += 3;
          exchange_register(saturn.C, saturn.R4, W_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
          dump_register(saturn.R4);
#endif
          return 0;
        default:
          return 1;
      }
    case 3:
      op3 = read_nibble(saturn.PC + 2);
      switch (op3) {
        case 0:		/* D0=A */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tD0=A\n", saturn.PC);
#endif
          saturn.PC += 3;
	  register_to_address(saturn.A, &saturn.D0, 0);
          return 0;
        case 1:		/* D1=A */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tD1=A\n", saturn.PC);
#endif
          saturn.PC += 3;
	  register_to_address(saturn.A, &saturn.D1, 0);
          return 0;
        case 2:		/* AD0EX */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tAD0EX\n", saturn.PC);
#endif
	  saturn.PC += 3;
	  exchange_reg(saturn.A, &saturn.D0, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.A);
	  dump_reg(saturn.D0);
#endif
	  return 0;
        case 3:		/* AD1EX */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tAD1EX\n", saturn.PC);
#endif
	  saturn.PC += 3;
	  exchange_reg(saturn.A, &saturn.D1, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.A);
	  dump_reg(saturn.D1);
#endif
	  return 0;
        case 4:		/* D0=C */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tD0=C\n", saturn.PC);
#endif
          saturn.PC += 3;
	  register_to_address(saturn.C, &saturn.D0, 0);
          return 0;
        case 5:		/* D1=C */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tD1=C\n", saturn.PC);
#endif
          saturn.PC += 3;
	  register_to_address(saturn.C, &saturn.D1, 0);
          return 0;
        case 6:		/* CD0EX */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tCD0EX\n", saturn.PC);
#endif
	  saturn.PC += 3;
	  exchange_reg(saturn.C, &saturn.D0, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.C);
	  dump_reg(saturn.D0);
#endif
	  return 0;
        case 7:		/* CD1EX */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tCD1EX\n", saturn.PC);
#endif
	  saturn.PC += 3;
	  exchange_reg(saturn.C, &saturn.D1, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.C);
	  dump_reg(saturn.D1);
#endif
	  return 0;
        case 8:		/* D0=AS */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tD0=AS\n", saturn.PC);
#endif
          saturn.PC += 3;
	  register_to_address(saturn.A, &saturn.D0, 1);
          return 0;
        case 9:		/* saturn.D1=AS */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tD1=AS\n", saturn.PC);
#endif
          saturn.PC += 3;
	  register_to_address(saturn.A, &saturn.D1, 1);
          return 0;
        case 0xa:	/* AD0XS */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tAD0XS\n", saturn.PC);
#endif
	  saturn.PC += 3;
	  exchange_reg(saturn.A, &saturn.D0, IN_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.A);
	  dump_reg(saturn.D0);
#endif
          return 0;
        case 0xb:	/* AD1XS */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tAD1XS\n", saturn.PC);
#endif
	  saturn.PC += 3;
	  exchange_reg(saturn.A, &saturn.D1, IN_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.A);
	  dump_reg(saturn.D1);
#endif
          return 0;
        case 0xc:	/* D0=CS */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tD0=CS\n", saturn.PC);
#endif
          saturn.PC += 3;
	  register_to_address(saturn.C, &saturn.D0, 1);
          return 0;
        case 0xd:	/* D1=CS */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tD1=CS\n", saturn.PC);
#endif
          saturn.PC += 3;
	  register_to_address(saturn.C, &saturn.D1, 1);
          return 0;
        case 0xe:	/* CD0XS */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tCD0XS\n", saturn.PC);
#endif
	  saturn.PC += 3;
	  exchange_reg(saturn.C, &saturn.D0, IN_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.C);
	  dump_reg(saturn.D0);
#endif
          return 0;
        case 0xf:	/* CD1XS */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tCD1XS\n", saturn.PC);
#endif
	  saturn.PC += 3;
	  exchange_reg(saturn.C, &saturn.D1, IN_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.C);
	  dump_reg(saturn.D1);
#endif
          return 0;
        default:
          return 1;
      }
    case 4:
      op3 = read_nibble(saturn.PC + 2);
      op = op3 < 8 ? 0xf : 6;
      switch(op3 & 7) {
        case 0:	/* DAT0=A */
	  saturn.PC += 3;
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tDAT0=A\t%s\n", saturn.PC, std_fields[op]);
#endif
	  store(saturn.D0, saturn.A, op);
          return 0;
	case 1:	/* DAT1=A */
	  saturn.PC += 3;
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tDAT1=A\t%s\n", saturn.PC, std_fields[op]);
#endif
	  store(saturn.D1, saturn.A, op);
          return 0;
	case 2:	/* A=DAT0 */
	  saturn.PC += 3;
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tA=DAT0\t%s\n", saturn.PC, std_fields[op]);
#endif
	  recall(saturn.A, saturn.D0, op);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
#endif
          return 0;
	case 3:	/* A=DAT1 */
	  saturn.PC += 3;
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tA=DAT1\t%s\n", saturn.PC, std_fields[op]);
#endif
	  recall(saturn.A, saturn.D1, op);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
#endif
          return 0;
	case 4:	/* DAT0=C */
	  saturn.PC += 3;
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tDAT0=C\t%s\n", saturn.PC, std_fields[op]);
#endif
	  store(saturn.D0, saturn.C, op);
          return 0;
	case 5:	/* DAT1=C */
	  saturn.PC += 3;
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tDAT1=C\t%s\n", saturn.PC, std_fields[op]);
#endif
	  store(saturn.D1, saturn.C, op);
          return 0;
	case 6:	/* C=DAT0 */
	  saturn.PC += 3;
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tC=DAT0\t%s\n", saturn.PC, std_fields[op]);
#endif
	  recall(saturn.C, saturn.D0, op);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
#endif
          return 0;
	case 7:	/* C=DAT1 */
	  saturn.PC += 3;
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tC=DAT1\t%s\n", saturn.PC, std_fields[op]);
#endif
	  recall(saturn.C, saturn.D1, op);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
#endif
          return 0;
	default:
	  return 1;
      }
    case 5:
      op3 = read_nibble(saturn.PC + 2);
      op4 = read_nibble(saturn.PC + 3);
      if (op3 >= 8) {
	switch(op3 & 7) {
          case 0:	/* DAT0=A */
	    saturn.PC += 4;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tDAT0=A\t%d\n", saturn.PC, op4+1);
#endif
	    store_n(saturn.D0, saturn.A, op4+1);
            return 0;
	  case 1:	/* DAT1=A */
	    saturn.PC += 4;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tDAT1=A\t%d\n", saturn.PC, op4+1);
#endif
	    store_n(saturn.D1, saturn.A, op4+1);
            return 0;
	  case 2:	/* A=DAT0 */
	    saturn.PC += 4;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tA=DAT0\t%d\n", saturn.PC, op4+1);
#endif
	    recall_n(saturn.A, saturn.D0, op4+1);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
#endif
            return 0;
	  case 3:	/* A=DAT1 */
	    saturn.PC += 4;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tA=DAT1\t%d\n", saturn.PC, op4+1);
#endif
	    recall_n(saturn.A, saturn.D1, op4+1);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
#endif
            return 0;
	  case 4:	/* DAT0=C */
	    saturn.PC += 4;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tDAT0=C\t%d\n", saturn.PC, op4+1);
#endif
	    store_n(saturn.D0, saturn.C, op4+1);
            return 0;
	  case 5:	/* DAT1=C */
	    saturn.PC += 4;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tDAT1=C\t%d\n", saturn.PC, op4+1);
#endif
	    store_n(saturn.D1, saturn.C, op4+1);
            return 0;
	  case 6:	/* C=DAT0 */
	    saturn.PC += 4;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tC=DAT0\t%d\n", saturn.PC, op4+1);
#endif
	    recall_n(saturn.C, saturn.D0, op4+1);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
#endif
            return 0;
	  case 7:	/* C=DAT1 */
	    saturn.PC += 4;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tC=DAT1\t%d\n", saturn.PC, op4+1);
#endif
	    recall_n(saturn.C, saturn.D1, op4+1);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
#endif
            return 0;
	  default:
	    return 1;
        }
      } else {
	switch(op3) {
          case 0:	/* DAT0=A */
	    saturn.PC += 4;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tDAT0=A\t%s\n", saturn.PC, std_fields[op4]);
#endif
	    store(saturn.D0, saturn.A, op4);
            return 0;
	  case 1:	/* DAT1=A */
	    saturn.PC += 4;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tDAT1=A\t%s\n", saturn.PC, std_fields[op4]);
#endif
	    store(saturn.D1, saturn.A, op4);
            return 0;
	  case 2:	/* A=DAT0 */
	    saturn.PC += 4;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tA=DAT0\t%s\n", saturn.PC, std_fields[op4]);
#endif
	    recall(saturn.A, saturn.D0, op4);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
#endif
            return 0;
	  case 3:	/* A=DAT1 */
	    saturn.PC += 4;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tA=DAT1\t%s\n", saturn.PC, std_fields[op4]);
#endif
	    recall(saturn.A, saturn.D1, op4);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
#endif
            return 0;
	  case 4:	/* DAT0=C */
	    saturn.PC += 4;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tDAT0=C\t%s\n", saturn.PC, std_fields[op4]);
#endif
	    store(saturn.D0, saturn.C, op4);
            return 0;
	  case 5:	/* DAT1=C */
	    saturn.PC += 4;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tDAT1=C\t%s\n", saturn.PC, std_fields[op4]);
#endif
	    store(saturn.D1, saturn.C, op4);
            return 0;
	  case 6:	/* C=DAT0 */
	    saturn.PC += 4;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tC=DAT0\t%s\n", saturn.PC, std_fields[op4]);
#endif
	    recall(saturn.C, saturn.D0, op4);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
#endif
            return 0;
	  case 7:	/* C=DAT1 */
	    saturn.PC += 4;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tC=DAT1\t%s\n", saturn.PC, std_fields[op4]);
#endif
	    recall(saturn.C, saturn.D1, op4);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
#endif
            return 0;
	  default:
	    return 1;
        }
      }
    case 6:
      op3 = read_nibble(saturn.PC + 2);
#ifdef DEBUG_PRINT
      dprintf("%.5lx\tD0=D0+\t%d\n", saturn.PC, op3+1);
#endif
      saturn.PC += 3;
      add_address(&saturn.D0, op3+1);
#ifdef DUMP_REGISTERS
      dump_reg(saturn.D0);
#endif
      return 0;
    case 7:
      op3 = read_nibble(saturn.PC + 2);
#ifdef DEBUG_PRINT
      dprintf("%.5lx\tD1=D1+\t%d\n", saturn.PC, op3+1);
#endif
      saturn.PC += 3;
      add_address(&saturn.D1, op3+1);
#ifdef DUMP_REGISTERS
      dump_reg(saturn.D1);
#endif
      return 0;
    case 8:
      op3 = read_nibble(saturn.PC + 2);
#ifdef DEBUG_PRINT
      dprintf("%.5lx\tD0=D0-\t%d\n", saturn.PC, op3+1);
#endif
      saturn.PC += 3;
      add_address(&saturn.D0, -(op3+1));
#ifdef DUMP_REGISTERS
      dump_reg(saturn.D0);
#endif
      return 0;
    case 9:
#ifdef DEBUG_PRINT
      op3 = read_nibble(saturn.PC + 2);
      dprintf("%.5lx\tD0=(2)\t%.2x\n", saturn.PC, op3);
#endif
      load_addr(&saturn.D0, saturn.PC+2, 2);
      saturn.PC += 4;
#ifdef DUMP_REGISTERS
      dump_reg(saturn.D0);
#endif
      return 0;
    case 0xa:
#ifdef DEBUG_PRINT
      op3 = read_nibble(saturn.PC + 2);
      dprintf("%.5lx\tD0=(4)\t%.4x\n", saturn.PC, op3);
#endif
      load_addr(&saturn.D0, saturn.PC+2, 4);
      saturn.PC += 6;
#ifdef DUMP_REGISTERS
      dump_reg(saturn.D0);
#endif
      return 0;
    case 0xb:
#ifdef DEBUG_PRINT
      op3 = read_nibble(saturn.PC + 2);
      dprintf("%.5lx\tD0=(5)\t%.5lx\n", saturn.PC, op3);
#endif
      load_addr(&saturn.D0, saturn.PC+2, 5);
      saturn.PC += 7;
#ifdef DUMP_REGISTERS
      dump_reg(saturn.D0);
#endif
      return 0;
    case 0xc:
      op3 = read_nibble(saturn.PC + 2);
#ifdef DEBUG_PRINT
      dprintf("%.5lx\tD1=D1-\t%d\n", saturn.PC, op3+1);
#endif
      saturn.PC += 3;
      add_address(&saturn.D1, -(op3+1));
#ifdef DUMP_REGISTERS
      dump_reg(saturn.D1);
#endif
      return 0;
    case 0xd:
#ifdef DEBUG_PRINT
      op3 = read_nibble(saturn.PC + 2);
      dprintf("%.5lx\tD1=(2)\t%.2x\n", saturn.PC, op3);
#endif
      load_addr(&saturn.D1, saturn.PC+2, 2);
      saturn.PC += 4;
#ifdef DUMP_REGISTERS
      dump_reg(saturn.D1);
#endif
      return 0;
    case 0xe:
#ifdef DEBUG_PRINT
      op3 = read_nibble(saturn.PC + 2);
      dprintf("%.5lx\tD1=(4)\t%.4x\n", saturn.PC, op3);
#endif
      load_addr(&saturn.D1, saturn.PC+2, 4);
      saturn.PC += 6;
#ifdef DUMP_REGISTERS
      dump_reg(saturn.D1);
#endif
      return 0;
    case 0xf:
#ifdef DEBUG_PRINT
      op3 = read_nibble(saturn.PC + 2);
      dprintf("%.5lx\tD1=(5)\t%.5lx\n", saturn.PC, op3);
#endif
      load_addr(&saturn.D1, saturn.PC+2, 5);
      saturn.PC += 7;
#ifdef DUMP_REGISTERS
      dump_reg(saturn.D1);
#endif
      return 0;
    default:
      return 1;
  }
}

int
#ifdef __FunctionProto__
decode_8_thru_f(long op1)
#else
decode_8_thru_f(op1)
long op1;
#endif
{
  long op2, op3, op4, op5, op6;

  op2 = read_nibble(saturn.PC + 1);

  switch (op1) {
    case 8:
      switch (op2) {
        case 0:
          return decode_group_80();
        case 1:
          op3 = read_nibble(saturn.PC + 2);
          switch (op3) {
            case 0:	/* ASLC */
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tASLC\n", saturn.PC);
#endif
              saturn.PC += 3;
              shift_left_circ_register(saturn.A, W_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.A);
#endif
              return 0;
            case 1:	/* BSLC */
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tBSLC\n", saturn.PC);
#endif
              saturn.PC += 3;
              shift_left_circ_register(saturn.B, W_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.B);
#endif
              return 0;
            case 2:	/* CSLC */
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tCSLC\n", saturn.PC);
#endif
              saturn.PC += 3;
              shift_left_circ_register(saturn.C, W_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.C);
#endif
              return 0;
            case 3:	/* DSLC */
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tDSLC\n", saturn.PC);
#endif
              saturn.PC += 3;
              shift_left_circ_register(saturn.D, W_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.D);
#endif
              return 0;
            case 4:	/* ASRC */
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tASRC\n", saturn.PC);
#endif
              saturn.PC += 3;
              shift_right_circ_register(saturn.A, W_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.A);
#endif
              return 0;
            case 5:	/* BSRC */
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tBSRC\n", saturn.PC);
#endif
              saturn.PC += 3;
              shift_right_circ_register(saturn.B, W_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.B);
#endif
              return 0;
            case 6:	/* CSRC */
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tCSRC\n", saturn.PC);
#endif
              saturn.PC += 3;
              shift_right_circ_register(saturn.C, W_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.C);
#endif
              return 0;
            case 7:	/* DSRC */
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tDSRC\n", saturn.PC);
#endif
              saturn.PC += 3;
              shift_right_circ_register(saturn.D, W_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.D);
#endif
              return 0;
            case 8:	/* R = R +/- CON */
              op4 = read_nibble(saturn.PC + 3);
              op5 = read_nibble(saturn.PC + 4);
              op6 = read_nibble(saturn.PC + 5);
              if (op5 < 8) {	/* PLUS */
                switch (op5 & 3) {
                  case 0:	/* A=A+CON */
#ifdef DEBUG_PRINT
                    dprintf("%.5lx\tA=A+CON\t%s,%d\n", saturn.PC,
                            std_fields[op4], op6+1);
#endif
                    saturn.PC += 6;
                    add_register_constant(saturn.A, op4, op6+1);
#ifdef DUMP_REGISTER
                    dump_register(saturn.A);
#endif
                    return 0;
                  case 1:	/* B=B+CON */
#ifdef DEBUG_PRINT
                    dprintf("%.5lx\tB=B+CON\t%s,%d\n", saturn.PC,
                            std_fields[op4], op6+1);
#endif
                    saturn.PC += 6;
                    add_register_constant(saturn.B, op4, op6+1);
#ifdef DUMP_REGISTER
                    dump_register(saturn.B);
#endif
                    return 0;
                  case 2:	/* C=C+CON */
#ifdef DEBUG_PRINT
                    dprintf("%.5lx\tC=C+CON\t%s,%d\n", saturn.PC,
                            std_fields[op4], op6+1);
#endif
                    saturn.PC += 6;
                    add_register_constant(saturn.C, op4, op6+1);
#ifdef DUMP_REGISTER
                    dump_register(saturn.C);
#endif
                    return 0;
                  case 3:	/* D=D+CON */
#ifdef DEBUG_PRINT
                    dprintf("%.5lx\tD=D+CON\t%s,%d\n", saturn.PC,
                            std_fields[op4], op6+1);
#endif
                    saturn.PC += 6;
                    add_register_constant(saturn.D, op4, op6+1);
#ifdef DUMP_REGISTER
                    dump_register(saturn.D);
#endif
                    return 0;
                  default:
                    return 1;
                }
              } else {		/* MINUS */
                switch (op5 & 3) {
                  case 0:	/* A=A-CON */
#ifdef DEBUG_PRINT
                    dprintf("%.5lx\tA=A-CON\t%s,%d\n", saturn.PC,
                            std_fields[op4], op6+1);
#endif
                    saturn.PC += 6;
                    sub_register_constant(saturn.A, op4, op6+1);
#ifdef DUMP_REGISTER
                    dump_register(saturn.A);
#endif
                    return 0;
                  case 1:	/* B=B-CON */
#ifdef DEBUG_PRINT
                    dprintf("%.5lx\tB=B-CON\t%s,%d\n", saturn.PC,
                            std_fields[op4], op6+1);
#endif
                    saturn.PC += 6;
                    sub_register_constant(saturn.B, op4, op6+1);
#ifdef DUMP_REGISTER
                    dump_register(saturn.B);
#endif
                    return 0;
                  case 2:	/* C=C-CON */
#ifdef DEBUG_PRINT
                    dprintf("%.5lx\tC=C-CON\t%s,%d\n", saturn.PC,
                            std_fields[op4], op6+1);
#endif
                    saturn.PC += 6;
                    sub_register_constant(saturn.C, op4, op6+1);
#ifdef DUMP_REGISTER
                    dump_register(saturn.C);
#endif
                    return 0;
                  case 3:	/* D=D-CON */
#ifdef DEBUG_PRINT
                    dprintf("%.5lx\tD=D-CON\t%s,%d\n", saturn.PC,
                            std_fields[op4], op6+1);
#endif
                    saturn.PC += 6;
                    sub_register_constant(saturn.D, op4, op6+1);
#ifdef DUMP_REGISTER
                    dump_register(saturn.D);
#endif
                    return 0;
                  default:
                    return 1;
                }
              }
            case 9:	/* R SRB FIELD */
              op4 = read_nibble(saturn.PC + 3);
              op5 = read_nibble(saturn.PC + 4);
              switch (op5 & 3) {
                case 0:
#ifdef DEBUG_PRINT
                  dprintf("%.5lx\tASRB.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                  saturn.PC += 5;
                  shift_right_bit_register(saturn.A, op4);
#ifdef DUMP_REGISTERS
                  dump_register(saturn.A);
#endif
                  return 0;
                case 1:
#ifdef DEBUG_PRINT
                  dprintf("%.5lx\tBSRB.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                  saturn.PC += 5;
                  shift_right_bit_register(saturn.B, op4);
#ifdef DUMP_REGISTERS
                  dump_register(saturn.B);
#endif
                  return 0;
                case 2:
#ifdef DEBUG_PRINT
                  dprintf("%.5lx\tCSRB.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                  saturn.PC += 5;
                  shift_right_bit_register(saturn.C, op4);
#ifdef DUMP_REGISTERS
                  dump_register(saturn.C);
#endif
                  return 0;
                case 3:
#ifdef DEBUG_PRINT
                  dprintf("%.5lx\tDSRB.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                  saturn.PC += 5;
                  shift_right_bit_register(saturn.D, op4);
#ifdef DUMP_REGISTERS
                  dump_register(saturn.D);
#endif
                  return 0;
                default:
                  return 1;
              }
            case 0xa:	/* R = R FIELD, etc. */
              op4 = read_nibble(saturn.PC + 3);
              op5 = read_nibble(saturn.PC + 4);
              op6 = read_nibble(saturn.PC + 5);
              switch (op5) {
                case 0:
                  switch (op6) {
                    case 0:		/* saturn.R0=A */
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tsaturn.R0=A.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.R0, saturn.A, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.R0);
#endif
                      return 0;
                    case 1:		/* saturn.R1=A */
                    case 5:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tsaturn.R1=A.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.R1, saturn.A, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.R1);
#endif
                      return 0;
                    case 2:		/* saturn.R2=A */
                    case 6:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tsaturn.R2=A.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.R2, saturn.A, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.R2);
#endif
                      return 0;
                    case 3:		/* saturn.R3=A */
                    case 7:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tsaturn.R3=A.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.R3, saturn.A, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.R3);
#endif
                      return 0;
                    case 4:		/* saturn.R4=A */
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tsaturn.R4=A.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.R4, saturn.A, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.R4);
#endif
                      return 0;
                    case 8:		/* saturn.R0=C */
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tsaturn.R0=C.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.R0, saturn.C, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.R0);
#endif
                      return 0;
                    case 9:		/* saturn.R1=C */
                    case 0xd:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tsaturn.R1=C.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.R1, saturn.C, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.R1);
#endif
                      return 0;
                    case 0xa:	/* saturn.R2=C */
                    case 0xe:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tsaturn.R2=C.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.R2, saturn.C, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.R2);
#endif
                      return 0;
                    case 0xb:	/* saturn.R3=C */
                    case 0xf:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tsaturn.R3=C.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.R3, saturn.C, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.R3);
#endif
                      return 0;
                    case 0xc:	/* saturn.R4=C */
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tsaturn.R4=C.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.R4, saturn.C, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.R4);
#endif
                      return 0;
                    default:
                      return 1;
                  }
                case 1:
                  switch (op6) {
                    case 0:		/* A=saturn.R0 */
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tA=saturn.R0.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.A, saturn.R0, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.A);
#endif
                      return 0;
                    case 1:		/* A=saturn.R1 */
                    case 5:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tA=saturn.R1.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.A, saturn.R1, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.A);
#endif
                      return 0;
                    case 2:		/* A=saturn.R2 */
                    case 6:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tA=saturn.R2.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.A, saturn.R2, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.A);
#endif
                      return 0;
                    case 3:		/* A=saturn.R3 */
                    case 7:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tA=saturn.R3.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.A, saturn.R3, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.A);
#endif
                      return 0;
                    case 4:		/* A=saturn.R4 */
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tA=saturn.R4.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.A, saturn.R4, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.A);
#endif
                      return 0;
                    case 8:		/* C=saturn.R0 */
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tC=saturn.R0.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.C, saturn.R0, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.C);
#endif
                      return 0;
                    case 9:		/* C=saturn.R1 */
                    case 0xd:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tC=saturn.R1.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.C, saturn.R1, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.C);
#endif
                      return 0;
                    case 0xa:	/* C=saturn.R2 */
                    case 0xe:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tC=saturn.R2.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.C, saturn.R2, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.C);
#endif
                      return 0;
                    case 0xb:	/* C=saturn.R3 */
                    case 0xf:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tC=saturn.R3.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.C, saturn.R3, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.C);
#endif
                      return 0;
                    case 0xc:	/* C=saturn.R4 */
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tC=saturn.R4.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      copy_register(saturn.C, saturn.R4, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.C);
#endif
                      return 0;
                    default:
                      return 1;
                  }
                case 2:
                  switch (op6) {
                    case 0:		/* Asaturn.R0EX */
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tAsaturn.R0EX.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      exchange_register(saturn.A, saturn.R0, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.A);
                      dump_register(saturn.R0);
#endif
                      return 0;
                    case 1:		/* Asaturn.R1EX */
                    case 5:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tAsaturn.R1EX.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      exchange_register(saturn.A, saturn.R1, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.A);
                      dump_register(saturn.R1);
#endif
                      return 0;
                    case 2:		/* Asaturn.R2EX */
                    case 6:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tAsaturn.R2EX.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      exchange_register(saturn.A, saturn.R2, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.A);
                      dump_register(saturn.R2);
#endif
                      return 0;
                    case 3:		/* Asaturn.R3EX */
                    case 7:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tAsaturn.R3EX.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      exchange_register(saturn.A, saturn.R3, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.A);
                      dump_register(saturn.R3);
#endif
                      return 0;
                    case 4:		/* Asaturn.R4EX */
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tAsaturn.R4EX.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      exchange_register(saturn.A, saturn.R4, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.A);
                      dump_register(saturn.R4);
#endif
                      return 0;
                    case 8:		/* Csaturn.R0EX */
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tCsaturn.R0EX.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      exchange_register(saturn.C, saturn.R0, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.C);
                      dump_register(saturn.R0);
#endif
                      return 0;
                    case 9:		/* Csaturn.R1EX */
                    case 0xd:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tCsaturn.R1EX.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      exchange_register(saturn.C, saturn.R1, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.C);
                      dump_register(saturn.R1);
#endif
                      return 0;
                    case 0xa:	/* Csaturn.R2EX */
                    case 0xe:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tCsaturn.R2EX.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      exchange_register(saturn.C, saturn.R2, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.C);
                      dump_register(saturn.R2);
#endif
                      return 0;
                    case 0xb:	/* Csaturn.R3EX */
                    case 0xf:
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tCsaturn.R3EX.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      exchange_register(saturn.C, saturn.R3, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.C);
                      dump_register(saturn.R3);
#endif
                      return 0;
                    case 0xc:	/* Csaturn.R4EX */
#ifdef DEBUG_PRINT
                      dprintf("%.5lx\tCsaturn.R4EX.F\t%s\n", saturn.PC, std_fields[op4]);
#endif
                      saturn.PC += 6;
                      exchange_register(saturn.C, saturn.R4, op4);
#ifdef DUMP_REGISTERS
                      dump_register(saturn.C);
                      dump_register(saturn.R4);
#endif
                      return 0;
                    default:
                      return 1;
                  }
                default:
                  return 1;
              }
            case 0xb:
              op4 = read_nibble(saturn.PC + 3);
              switch(op4) {
                case 2:		/* PC=A */
#ifdef DEBUG_PRINT
                  dprintf("%.5lx\tPC=A\n", saturn.PC);
#endif
                  jumpaddr = dat_to_addr(saturn.A);
                  saturn.PC = jumpaddr;
#ifdef DEBUG_PRINT
		  dprintf("PC = %.5lx\n", saturn.PC);
#endif
                  return 0;
                case 3:		/* PC=C */
#ifdef DEBUG_PRINT
                  dprintf("%.5lx\tPC=C\n", saturn.PC);
#endif
                  jumpaddr = dat_to_addr(saturn.C);
                  saturn.PC = jumpaddr;
                  return 0;
                case 4:		/* A=PC */
#ifdef DEBUG_PRINT
                  dprintf("%.5lx\tA=PC\n", saturn.PC);
#endif
                  saturn.PC += 4;
                  addr_to_dat(saturn.PC, saturn.A);
#ifdef DUMP_REGISTERS
                  dump_register(saturn.A);
#endif
                  return 0;
                case 5:		/* C=PC */
#ifdef DEBUG_PRINT
                  dprintf("%.5lx\tC=PC\n", saturn.PC);
#endif
                  saturn.PC += 4;
                  addr_to_dat(saturn.PC, saturn.C);
#ifdef DUMP_REGISTERS
                  dump_register(saturn.C);
#endif
                  return 0;
                case 6:		/* APCEX */
#ifdef DEBUG_PRINT
                  dprintf("%.5lx\tAPCEX\n", saturn.PC);
#endif
		  saturn.PC += 4;
                  jumpaddr = dat_to_addr(saturn.A);
                  addr_to_dat(saturn.PC, saturn.A);
                  saturn.PC = jumpaddr;
#ifdef DUMP_REGISTERS
                  dump_register(saturn.A);
#endif
                  return 0;
                case 7: 	/* CPCEX */
#ifdef DEBUG_PRINT
                  dprintf("%.5lx\tCPCEX\n", saturn.PC);
#endif
		  saturn.PC += 4;
                  jumpaddr = dat_to_addr(saturn.C);
                  addr_to_dat(saturn.PC, saturn.C);
                  saturn.PC = jumpaddr;
#ifdef DUMP_REGISTERS
                  dump_register(saturn.C);
#endif
                  return 0;
                default:
                  return 1;
              }
            case 0xc:	/* ASRB */
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tASRB\n", saturn.PC);
#endif
              saturn.PC += 3;
              shift_right_bit_register(saturn.A, W_FIELD);
#ifdef DUMP_REGISTERs
              dump_register(saturn.A);
#endif
              return 0;
            case 0xd:	/* BSRB */
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tBSRB\n", saturn.PC);
#endif
              saturn.PC += 3;
              shift_right_bit_register(saturn.B, W_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.B);
#endif
              return 0;
            case 0xe:	/* CSRB */
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tCSRB\n", saturn.PC);
#endif
              saturn.PC += 3;
              shift_right_bit_register(saturn.C, W_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.C);
#endif
              return 0;
            case 0xf:	/* DSRB */
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tDSRB\n", saturn.PC);
#endif
              saturn.PC += 3;
              shift_right_bit_register(saturn.D, W_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.D);
#endif
              return 0;
            default:
              return 1;
          }
        case 2:
          op3 = read_nibble(saturn.PC + 2);
          switch (op3) {
            case 1:
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tXM=0\n", saturn.PC);
#endif
              break;
            case 2:
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tSB=0\n", saturn.PC);
#endif
              break;
            case 4:
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tSR=0\n", saturn.PC);
#endif
              break;
            case 8:
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tMP=0\n", saturn.PC);
#endif
              break;
            case 0xf:
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tCLRHST\n", saturn.PC);
#endif
              break;
            default:
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tHS=0\t%d\n", saturn.PC, op3);
#endif
              break;
          }
          saturn.PC += 3;
          clear_hardware_stat(op3);
          return 0;
        case 3:
          op3 = read_nibble(saturn.PC + 2);
          switch (op3) {
            case 1:
#ifdef DEBUG_PRINT
              dprintf("%.5lx\t?XM=0\n", saturn.PC);
#endif
              break;
            case 2:
#ifdef DEBUG_PRINT
              dprintf("%.5lx\t?SB=0\n", saturn.PC);
#endif
              break;
            case 4:
#ifdef DEBUG_PRINT
              dprintf("%.5lx\t?SR=0\n", saturn.PC);
#endif
              break;
            case 8:
#ifdef DEBUG_PRINT
              dprintf("%.5lx\t?MP=0\n", saturn.PC);
#endif
              break;
            default:
#ifdef DEBUG_PRINT
              dprintf("%.5lx\t?HS=0\t%d\n", saturn.PC, op3);
#endif
              break;
          }
          saturn.CARRY = is_zero_hardware_stat(op3);
	  if (saturn.CARRY) {
	    saturn.PC += 3;
            op4 = read_nibbles(saturn.PC, 2);
      	    if (op4) {
              if (op4 & 0x80)
                op4 |= 0xffffff00;
              jumpaddr = (saturn.PC + op4) & 0xfffff;
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tGOYES\t%.5lx\n", saturn.PC, jumpaddr);
#endif
	      saturn.PC = jumpaddr;
            } else {
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tRTNYES\n", saturn.PC);
#endif
	      saturn.PC = pop_return_addr();
            }
	  } else {
            saturn.PC += 5;
	  }
          return 0;
        case 4:
        case 5:
          op3 = read_nibble(saturn.PC + 2);
	  if (op2 == 4) {
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tST=0\t%d\n", saturn.PC, op3);
#endif
	    saturn.PC += 3;
            clear_program_stat(op3);
          } else {
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tST=1\t%d\n", saturn.PC, op3);
#endif
	    saturn.PC += 3;
	    set_program_stat(op3);
          }
          return 0;
        case 6:
        case 7:
          op3 = read_nibble(saturn.PC + 2);
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\t?ST=%d\t%d\n", saturn.PC, (op2==6)?0:1, op3);
#endif
	  if (op2 == 6)
            saturn.CARRY = (get_program_stat(op3) == 0)?1:0;
          else
            saturn.CARRY = (get_program_stat(op3) != 0)?1:0;
	  if (saturn.CARRY) {
	    saturn.PC += 3;
            op4 = read_nibbles(saturn.PC, 2);
      	    if (op4) {
              if (op4 & 0x80)
                op4 |= 0xffffff00;
              jumpaddr = (saturn.PC + op4) & 0xfffff;
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tGOYES\t%.5lx\n", saturn.PC, jumpaddr);
#endif
	      saturn.PC = jumpaddr;
            } else {
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tRTNYES\n", saturn.PC);
#endif
	      saturn.PC = pop_return_addr();
            }
	  } else {
            saturn.PC += 5;
	  }
          return 0;
        case 8:
        case 9:
          op3 = read_nibble(saturn.PC + 2);
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\t?P%s\t%d\n", saturn.PC, (op2==8)?"#":"=", op3);
#endif
	  if (op2 == 8)
            saturn.CARRY = (saturn.P != op3)?1:0;
          else
            saturn.CARRY = (saturn.P == op3)?1:0;
	  if (saturn.CARRY) {
	    saturn.PC += 3;
            op4 = read_nibbles(saturn.PC, 2);
      	    if (op4) {
              if (op4 & 0x80)
                op4 |= 0xffffff00;
              jumpaddr = (saturn.PC + op4) & 0xfffff;
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tGOYES\t%.5lx\n", saturn.PC, jumpaddr);
#endif
	      saturn.PC = jumpaddr;
            } else {
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tRTNYES\n", saturn.PC);
#endif
	      saturn.PC = pop_return_addr();
            }
	  } else {
            saturn.PC += 5;
	  }
          return 0;
        case 0xa:
          op3 = read_nibble(saturn.PC + 2);
          switch(op3) {
            case 0:	/* ?A=B */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?A=B\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_equal_register(saturn.A, saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.A);
              dump_register(saturn.B);
#endif
              break;
            case 1:	/* ?B=C */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?B=C\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_equal_register(saturn.B, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.B);
              dump_register(saturn.C);
#endif
              break;
            case 2:	/* ?A=C */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?A=C\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_equal_register(saturn.A, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.A);
              dump_register(saturn.C);
#endif
              break;
            case 3:	/* ?C=D */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?C=D\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_equal_register(saturn.C, saturn.D, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.C);
              dump_register(saturn.D);
#endif
              break;
            case 4:	/* ?A#B */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?A#B\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_not_equal_register(saturn.A, saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.A);
              dump_register(saturn.B);
#endif
              break;
            case 5:	/* ?B#C */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?B#C\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_not_equal_register(saturn.B, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.B);
              dump_register(saturn.C);
#endif
              break;
            case 6:	/* ?A#C */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?A#C\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_not_equal_register(saturn.A, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.A);
              dump_register(saturn.C);
#endif
              break;
            case 7:	/* ?C#D */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?C#D\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_not_equal_register(saturn.C, saturn.D, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.C);
              dump_register(saturn.D);
#endif
              break;
            case 8:	/* ?A=0 */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?A=0\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_zero_register(saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.A);
#endif
              break;
            case 9:	/* ?B=0 */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?B=0\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_zero_register(saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.B);
#endif
              break;
            case 0xa:	/* ?C=0 */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?C=0\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_zero_register(saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.C);
#endif
              break;
            case 0xb:	/* ?D=0 */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?D=0\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_zero_register(saturn.D, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.D);
#endif
              break;
            case 0xc:	/* ?A#0 */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?A#0\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_not_zero_register(saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.A);
#endif
              break;
            case 0xd:	/* ?B#0 */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?B#0\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_not_zero_register(saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.B);
#endif
              break;
            case 0xe:	/* ?C#0 */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?C#0\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_not_zero_register(saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.C);
#endif
              break;
            case 0xf:	/* ?D#0 */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?D#0\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_not_zero_register(saturn.D, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.D);
#endif
              break;
            default:
              return 1;
          }
	  if (saturn.CARRY) {
	    saturn.PC += 3;
            op4 = read_nibbles(saturn.PC, 2);
      	    if (op4) {
              if (op4 & 0x80)
                op4 |= 0xffffff00;
              jumpaddr = (saturn.PC + op4) & 0xfffff;
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tGOYES\t%.5lx\n", saturn.PC, jumpaddr);
#endif
	      saturn.PC = jumpaddr;
            } else {
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tRTNYES\n", saturn.PC);
#endif
	      saturn.PC = pop_return_addr();
            }
	  } else {
            saturn.PC += 5;
	  }
          return 0;
	case 0xb:
          op3 = read_nibble(saturn.PC + 2);
          switch (op3) {
            case 0:	/* ?A>B */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?A>B\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_greater_register(saturn.A, saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.A);
              dump_register(saturn.B);
#endif
              break;
            case 1:	/* ?B>C */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?B>C\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_greater_register(saturn.B, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.B);
              dump_register(saturn.C);
#endif
              break;
            case 2:	/* ?C>A */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?C>A\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_greater_register(saturn.C, saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.C);
              dump_register(saturn.A);
#endif
              break;
            case 3:	/* ?D>C */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?D>C\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_greater_register(saturn.D, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.D);
              dump_register(saturn.C);
#endif
              break;
            case 4:	/* ?A<B */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?A<B\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_less_register(saturn.A, saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.A);
              dump_register(saturn.B);
#endif
              break;
            case 5:	/* ?B<C */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?B<C\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_less_register(saturn.B, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.B);
              dump_register(saturn.C);
#endif
              break;
            case 6:	/* ?C<A */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?C<A\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_less_register(saturn.C, saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.C);
              dump_register(saturn.A);
#endif
              break;
            case 7:	/* ?D<C */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?D<C\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_less_register(saturn.D, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.D);
              dump_register(saturn.C);
#endif
              break;
            case 8:	/* ?A>=B */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?A>=B\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_greater_or_equal_register(saturn.A, saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.A);
              dump_register(saturn.B);
#endif
              break;
            case 9:	/* ?B>=C */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?B>=C\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_greater_or_equal_register(saturn.B, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.B);
              dump_register(saturn.C);
#endif
              break;
            case 0xa:	/* ?C>=A */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?C>=A\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_greater_or_equal_register(saturn.C, saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.C);
              dump_register(saturn.A);
#endif
              break;
            case 0xb:	/* ?D>=C */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?D>=C\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_greater_or_equal_register(saturn.D, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.D);
              dump_register(saturn.C);
#endif
              break;
            case 0xc:	/* ?A<=B */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?A<=B\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_less_or_equal_register(saturn.A, saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.A);
              dump_register(saturn.B);
#endif
              break;
            case 0xd:	/* ?B<=C */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?B<=C\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_less_or_equal_register(saturn.B, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.B);
              dump_register(saturn.C);
#endif
              break;
            case 0xe:	/* ?C<=A */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?C<=A\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_less_or_equal_register(saturn.C, saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.C);
              dump_register(saturn.A);
#endif
              break;
            case 0xf:	/* ?D<=C */
#ifdef DEBUG_PRINT
	      dprintf("%.5lx\t?D<=C\tA\n", saturn.PC);
#endif
	      saturn.CARRY = is_less_or_equal_register(saturn.D, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
              dump_register(saturn.D);
              dump_register(saturn.C);
#endif
              break;
            default:
              return 1;
          }
	  if (saturn.CARRY) {
	    saturn.PC += 3;
            op4 = read_nibbles(saturn.PC, 2);
      	    if (op4) {
              if (op4 & 0x80)
                op4 |= 0xffffff00;
              jumpaddr = (saturn.PC + op4) & 0xfffff;
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tGOYES\t%.5lx\n", saturn.PC, jumpaddr);
#endif
	      saturn.PC = jumpaddr;
            } else {
#ifdef DEBUG_PRINT
              dprintf("%.5lx\tRTNYES\n", saturn.PC);
#endif
	      saturn.PC = pop_return_addr();
            }
	  } else {
            saturn.PC += 5;
	  }
          return 0;
	case 0xc:
          op3 = read_nibbles(saturn.PC + 2, 4);
          if (op3 & 0x8000)
            op3 |= 0xffff0000;
          jumpaddr = (saturn.PC + op3 + 2) & 0xfffff;
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tGOLONG\t%.5lx\n", saturn.PC, jumpaddr);
#endif
          saturn.PC = jumpaddr;
	  return 0;
	case 0xd:
          op3 = read_nibbles(saturn.PC + 2, 5);
          jumpaddr = op3;
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tGOVLNG\t%.5lx\n", saturn.PC, jumpaddr);
#endif
          saturn.PC = jumpaddr;
	  return 0;
	case 0xe:
          op3 = read_nibbles(saturn.PC + 2, 4);
          if (op3 & 0x8000)
            op3 |= 0xffff0000;
          jumpaddr = (saturn.PC + op3 + 6) & 0xfffff;
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tGOSUBL\t%.5lx\n", saturn.PC, jumpaddr);
#endif
	  push_return_addr(saturn.PC + 6);
          saturn.PC = jumpaddr;
	  return 0;
	case 0xf:
          op3 = read_nibbles(saturn.PC + 2, 5);
          jumpaddr = op3;
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tGOSBVL\t%.5lx\n", saturn.PC, jumpaddr);
#endif
	  push_return_addr(saturn.PC + 7);
          saturn.PC = jumpaddr;
	  return 0;
        default:
          return 1;
      }
    case 9:
      op3 = read_nibble(saturn.PC + 2);
      if (op2 < 8) {
        switch(op3) {
          case 0:	/* ?A=B */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?A=B\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_equal_register(saturn.A, saturn.B, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
            dump_register(saturn.B);
#endif
            break;
          case 1:	/* ?B=C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?B=C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_equal_register(saturn.B, saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.B);
            dump_register(saturn.C);
#endif
            break;
          case 2:	/* ?A=C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?A=C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_equal_register(saturn.A, saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
            dump_register(saturn.C);
#endif
            break;
          case 3:	/* ?C=D */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?C=D\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_equal_register(saturn.C, saturn.D, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
            dump_register(saturn.D);
#endif
            break;
          case 4:	/* ?A#B */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?A#B\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_not_equal_register(saturn.A, saturn.B, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
            dump_register(saturn.B);
#endif
            break;
          case 5:	/* ?B#C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?B#C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_not_equal_register(saturn.B, saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.B);
            dump_register(saturn.C);
#endif
            break;
          case 6:	/* ?A#C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?A#C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_not_equal_register(saturn.A, saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
            dump_register(saturn.C);
#endif
            break;
          case 7:	/* ?C#D */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?C#D\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_not_equal_register(saturn.C, saturn.D, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
            dump_register(saturn.D);
#endif
            break;
          case 8:	/* ?A=0 */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?A=0\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_zero_register(saturn.A, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
#endif
            break;
          case 9:	/* ?B=0 */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?B=0\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_zero_register(saturn.B, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.B);
#endif
            break;
          case 0xa:	/* ?C=0 */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?C=0\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_zero_register(saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
#endif
            break;
          case 0xb:	/* ?D=0 */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?D=0\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_zero_register(saturn.D, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.D);
#endif
            break;
          case 0xc:	/* ?A#0 */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?A#0\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_not_zero_register(saturn.A, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
#endif
            break;
          case 0xd:	/* ?B#0 */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?B#0\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_not_zero_register(saturn.B, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.B);
#endif
            break;
          case 0xe:	/* ?C#0 */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?C#0\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_not_zero_register(saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
#endif
            break;
          case 0xf:	/* ?D#0 */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?D#0\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.CARRY = is_not_zero_register(saturn.D, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.D);
#endif
            break;
          default:
            return 1;
        }
      } else {
        op2 &= 7;
        switch (op3) {
          case 0:	/* ?A>B */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?A>B\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_greater_register(saturn.A, saturn.B, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
            dump_register(saturn.B);
#endif
            break;
          case 1:	/* ?B>C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?B>C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_greater_register(saturn.B, saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.B);
            dump_register(saturn.C);
#endif
            break;
          case 2:	/* ?C>A */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?C>A\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_greater_register(saturn.C, saturn.A, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
            dump_register(saturn.A);
#endif
            break;
          case 3:	/* ?D>C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?D>C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_greater_register(saturn.D, saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.D);
            dump_register(saturn.C);
#endif
            break;
          case 4:	/* ?A<B */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?A<B\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_less_register(saturn.A, saturn.B, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
            dump_register(saturn.B);
#endif
            break;
          case 5:	/* ?B<C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?B<C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_less_register(saturn.B, saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.B);
            dump_register(saturn.C);
#endif
            break;
          case 6:	/* ?C<A */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?C<A\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_less_register(saturn.C, saturn.A, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
            dump_register(saturn.A);
#endif
            break;
          case 7:	/* ?D<C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?D<C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_less_register(saturn.D, saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.D);
            dump_register(saturn.C);
#endif
            break;
          case 8:	/* ?A>=B */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?A>=B\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_greater_or_equal_register(saturn.A, saturn.B, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
            dump_register(saturn.B);
#endif
            break;
          case 9:	/* ?B>=C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?B>=C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_greater_or_equal_register(saturn.B, saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.B);
            dump_register(saturn.C);
#endif
            break;
          case 0xa:	/* ?C>=A */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?C>=A\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_greater_or_equal_register(saturn.C, saturn.A, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
            dump_register(saturn.A);
#endif
            break;
          case 0xb:	/* ?D>=C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?D>=C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_greater_or_equal_register(saturn.D, saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.D);
            dump_register(saturn.C);
#endif
            break;
          case 0xc:	/* ?A<=B */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?A<=B\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_less_or_equal_register(saturn.A, saturn.B, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
            dump_register(saturn.B);
#endif
            break;
          case 0xd:	/* ?B<=C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?B<=C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_less_or_equal_register(saturn.B, saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.B);
            dump_register(saturn.C);
#endif
            break;
          case 0xe:	/* ?C<=A */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?C<=A\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_less_or_equal_register(saturn.C, saturn.A, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
            dump_register(saturn.A);
#endif
            break;
          case 0xf:	/* ?D<=C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\t?D<=C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.CARRY = is_less_or_equal_register(saturn.D, saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.D);
            dump_register(saturn.C);
#endif
            break;
          default:
            return 1;
        }
      }
      if (saturn.CARRY) {
	saturn.PC += 3;
        op4 = read_nibbles(saturn.PC, 2);
      	if (op4) {
          if (op4 & 0x80)
            op4 |= 0xffffff00;
          jumpaddr = (saturn.PC + op4) & 0xfffff;
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tGOYES\t%.5lx\n", saturn.PC, jumpaddr);
#endif
	  saturn.PC = jumpaddr;
        } else {
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tRTNYES\n", saturn.PC);
#endif
	  saturn.PC = pop_return_addr();
        }
      } else {
        saturn.PC += 5;
      }
      return 0;
    case 0xa:
      op3 = read_nibble(saturn.PC + 2);
      if (op2 < 8) {
	switch(op3) {
	  case 0:	/* A=A+B */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tA=A+B\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    add_register(saturn.A, saturn.A, saturn.B, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.A);
#endif
	    return 0;
	  case 1:	/* B=B+C */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tB=B+C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    add_register(saturn.B, saturn.B, saturn.C, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.B);
#endif
	    return 0;
	  case 2:	/* C=C+A */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tC=C+A\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    add_register(saturn.C, saturn.C, saturn.A, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.C);
#endif
	    return 0;
	  case 3:	/* D=D+C */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tD=D+C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    add_register(saturn.D, saturn.D, saturn.C, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.D);
#endif
	    return 0;
	  case 4:	/* A=A+A */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tA=A+A\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    add_register(saturn.A, saturn.A, saturn.A, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.A);
#endif
	    return 0;
	  case 5:	/* B=B+B */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tB=B+B\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    add_register(saturn.B, saturn.B, saturn.B, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.B);
#endif
	    return 0;
	  case 6:	/* C=C+C */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tC=C+C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    add_register(saturn.C, saturn.C, saturn.C, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.C);
#endif
	    return 0;
	  case 7:	/* D=D+D */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tD=D+D\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    add_register(saturn.D, saturn.D, saturn.D, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.D);
#endif
	    return 0;
	  case 8:	/* B=B+A */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tB=B+A\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    add_register(saturn.B, saturn.B, saturn.A, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.B);
#endif
	    return 0;
	  case 9:	/* C=C+B */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tC=C+B\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    add_register(saturn.C, saturn.C, saturn.B, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.C);
#endif
	    return 0;
	  case 0xa:	/* A=A+C */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tA=A+C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    add_register(saturn.A, saturn.A, saturn.C, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.A);
#endif
	    return 0;
	  case 0xb:	/* C=C+D */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tC=C+D\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    add_register(saturn.C, saturn.C, saturn.D, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.C);
#endif
	    return 0;
	  case 0xc:	/* A=A-1 */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tA=A-1\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    dec_register(saturn.A, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.A);
#endif
	    return 0;
	  case 0xd:	/* B=B-1 */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tB=B-1\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    dec_register(saturn.B, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.B);
#endif
	    return 0;
	  case 0xe:	/* C=C-1 */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tC=C-1\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    dec_register(saturn.C, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.C);
#endif
	    return 0;
	  case 0xf:	/* D=D-1 */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tD=D-1\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    dec_register(saturn.D, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.D);
#endif
	    return 0;
	  default:
	    return 1;
	}
      } else {
	op2 &= 7;
	switch(op3) {
	  case 0:	/* A=0 */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tA=0\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    zero_register(saturn.A, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
#endif
	    return 0;
	  case 1:	/* B=0 */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tB=0\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    zero_register(saturn.B, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.B);
#endif
	    return 0;
	  case 2:	/* C=0 */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tC=0\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    zero_register(saturn.C, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.C);
#endif
	    return 0;
	  case 3:	/* D=0 */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tD=0\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    zero_register(saturn.D, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.D);
#endif
	    return 0;
	  case 4:	/* A=B */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tA=B\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    copy_register(saturn.A, saturn.B, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.A);
#endif
	    return 0;
	  case 5:	/* B=C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tB=C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    copy_register(saturn.B, saturn.C, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.B);
#endif
	    return 0;
	  case 6:	/* C=A */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tC=A\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    copy_register(saturn.C, saturn.A, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.C);
#endif
	    return 0;
	  case 7:	/* D=C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tD=C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    copy_register(saturn.D, saturn.C, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.D);
#endif
	    return 0;
	  case 8:	/* B=A */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tB=A\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    copy_register(saturn.B, saturn.A, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.B);
#endif
	    return 0;
	  case 9:	/* C=B */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tC=B\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    copy_register(saturn.C, saturn.B, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.C);
#endif
	    return 0;
	  case 0xa:	/* A=C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tA=C\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    copy_register(saturn.A, saturn.C, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.A);
#endif
	    return 0;
	  case 0xb:	/* C=D */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tC=D\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    copy_register(saturn.C, saturn.D, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.C);
#endif
	    return 0;
	  case 0xc:	/* ABEX */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tABEX\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    exchange_register(saturn.A, saturn.B, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.A);
	    dump_register(saturn.B);
#endif
	    return 0;
	  case 0xd:	/* BCEX */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tBCEX\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    exchange_register(saturn.B, saturn.C, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.B);
	    dump_register(saturn.C);
#endif
	    return 0;
	  case 0xe:	/* ACEX */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tACEX\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    exchange_register(saturn.A, saturn.C, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.A);
	    dump_register(saturn.C);
#endif
	    return 0;
	  case 0xf:	/* CDEX */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tCDEX\t%s\n", saturn.PC, std_fields[op2]);
#endif
	    saturn.PC += 3;
	    exchange_register(saturn.C, saturn.D, op2);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.C);
	    dump_register(saturn.D);
#endif
	    return 0;
	  default:
	    return 1;
	}
      }
    case 0xb:
      op3 = read_nibble(saturn.PC + 2);
      if (op2 < 8) {
        switch (op3) {
          case 0:	/* A=A-B */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tA=A-B\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            sub_register(saturn.A, saturn.A, saturn.B, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
#endif
            return 0;
          case 1:	/* B=B-C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tB=B-C\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            sub_register(saturn.B, saturn.B, saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.B);
#endif
            return 0;
          case 2:	/* C=C-A */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tC=C-A\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            sub_register(saturn.C, saturn.C, saturn.A, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
#endif
            return 0;
          case 3:	/* D=D-C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tD=D-C\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            sub_register(saturn.D, saturn.D, saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.D);
#endif
            return 0;
          case 4:	/* A=A+1 */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tA=A+1\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
	    inc_register(saturn.A, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
#endif
            return 0;
          case 5:	/* B=B+1 */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tB=B+1\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
	    inc_register(saturn.B, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.B);
#endif
            return 0;
          case 6:	/* C=C+1 */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tC=C+1\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
	    inc_register(saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
#endif
            return 0;
          case 7:	/* D=D+1 */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tD=D+1\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
	    inc_register(saturn.D, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.D);
#endif
            return 0;
          case 8:	/* B=B-A */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tB=B-A\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            sub_register(saturn.B, saturn.B, saturn.A, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.B);
#endif
            return 0;
          case 9:	/* C=C-B */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tC=C-B\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            sub_register(saturn.C, saturn.C, saturn.B, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
#endif
            return 0;
          case 0xa:	/* A=A-C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tA=A-C\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            sub_register(saturn.A, saturn.A, saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
#endif
            return 0;
          case 0xb:	/* C=C-D */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tC=C-D\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            sub_register(saturn.C, saturn.C, saturn.D, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
#endif
            return 0;
          case 0xc:	/* A=B-A */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tA=B-A\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            sub_register(saturn.A, saturn.B, saturn.A, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
#endif
            return 0;
          case 0xd:	/* B=C-B */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tB=C-B\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            sub_register(saturn.B, saturn.C, saturn.B, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.B);
#endif
            return 0;
          case 0xe:	/* C=A-C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tC=A-C\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            sub_register(saturn.C, saturn.A, saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
#endif
            return 0;
          case 0xf:	/* D=C-D */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tD=C-D\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            sub_register(saturn.D, saturn.C, saturn.D, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.D);
#endif
            return 0;
          default:
            return 1;
        }
      } else {
	op2 &= 7;
        switch (op3) {
          case 0:	/* ASL */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tASL\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            shift_left_register(saturn.A, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
#endif
            return 0;
          case 1:	/* BSL */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tBSL\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            shift_left_register(saturn.B, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.B);
#endif
            return 0;
          case 2:	/* CSL */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tCSL\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            shift_left_register(saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
#endif
            return 0;
          case 3:	/* DSL */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tDSL\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            shift_left_register(saturn.D, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.D);
#endif
            return 0;
          case 4:	/* ASR */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tASR\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            shift_right_register(saturn.A, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
#endif
            return 0;
          case 5:	/* BSR */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tBSR\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            shift_right_register(saturn.B, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.B);
#endif
            return 0;
          case 6:	/* CSR */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tCSR\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            shift_right_register(saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
#endif
            return 0;
          case 7:	/* DSR */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tDSR\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
            shift_right_register(saturn.D, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.D);
#endif
            return 0;
          case 8:	/* A=-A */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tA=-A\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
	    complement_2_register(saturn.A, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
#endif
            return 0;
          case 9:	/* B=-B */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tB=-B\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
	    complement_2_register(saturn.B, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.B);
#endif
            return 0;
          case 0xa:	/* C=-C */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tC=-C\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
	    complement_2_register(saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
#endif
            return 0;
          case 0xb:	/* D=-D */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tD=-D\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
	    complement_2_register(saturn.D, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.D);
#endif
            return 0;
          case 0xc:	/* A=-A-1 */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tA=-A-1\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
	    complement_1_register(saturn.A, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.A);
#endif
            return 0;
          case 0xd:	/* B=-B-1 */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tB=-B-1\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
	    complement_1_register(saturn.B, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.B);
#endif
            return 0;
          case 0xe:	/* C=-C-1 */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tC=-C-1\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
	    complement_1_register(saturn.C, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.C);
#endif
            return 0;
          case 0xf:	/* D=-D-1 */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tD=-D-1\t%s\n", saturn.PC, std_fields[op2]);
#endif
            saturn.PC += 3;
	    complement_1_register(saturn.D, op2);
#ifdef DUMP_REGISTERS
            dump_register(saturn.D);
#endif
            return 0;
          default:
            return 1;
        }
      }
    case 0xc:
      switch(op2) {
        case 0:		/* A=A+B */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tA=A+B\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  add_register(saturn.A, saturn.A, saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.A);
#endif
          return 0;
        case 1:		/* B=B+C */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tB=B+C\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  add_register(saturn.B, saturn.B, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.B);
#endif
          return 0;
        case 2:		/* C=C+A */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tC=C+A\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  add_register(saturn.C, saturn.C, saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.C);
#endif
          return 0;
        case 3:		/* D=D+C */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tD=D+C\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  add_register(saturn.D, saturn.D, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.D);
#endif
          return 0;
        case 4:		/* A=A+A */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tA=A+A\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  add_register(saturn.A, saturn.A, saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.A);
#endif
          return 0;
        case 5:		/* B=B+B */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tB=B+B\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  add_register(saturn.B, saturn.B, saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.B);
#endif
          return 0;
        case 6:		/* C=C+C */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tC=C+C\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  add_register(saturn.C, saturn.C, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.C);
#endif
          return 0;
        case 7:		/* D=D+D */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tD=D+D\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  add_register(saturn.D, saturn.D, saturn.D, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.D);
#endif
          return 0;
        case 8:		/* B=B+A */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tB=B+A\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  add_register(saturn.B, saturn.B, saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.B);
#endif
          return 0;
        case 9:		/* C=C+B */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tC=C+B\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  add_register(saturn.C, saturn.C, saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.C);
#endif
          return 0;
        case 0xa:	/* A=A+C */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tA=A+C\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  add_register(saturn.A, saturn.A, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.A);
#endif
          return 0;
        case 0xb:	/* C=C+D */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tC=C+D\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  add_register(saturn.C, saturn.C, saturn.D, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.C);
#endif
          return 0;
        case 0xc:	/* A=A-1 */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tA=A-1\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  dec_register(saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.A);
#endif
	  return 0;
        case 0xd:	/* B=B-1 */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tB=B-1\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  dec_register(saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.B);
#endif
	  return 0;
        case 0xe:	/* C=C-1 */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tC=C-1\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  dec_register(saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.C);
#endif
	  return 0;
        case 0xf:	/* D=D-1 */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tD=D-1\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  dec_register(saturn.D, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.D);
#endif
	  return 0;
        default:
          return 1;
      }
    case 0xd:
      switch(op2) {
	case 0:		/* A=0 */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tA=0\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  zero_register(saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.A);
#endif
	  return 0;
	case 1:		/* B=0 */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tB=0\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  zero_register(saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.B);
#endif
	  return 0;
	case 2:		/* C=0 */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tC=0\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  zero_register(saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.C);
#endif
	  return 0;
	case 3:		/* D=0 */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tD=0\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  zero_register(saturn.D, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.D);
#endif
	  return 0;
        case 4:		/* A=B */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tA=B\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  copy_register(saturn.A, saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.A);
#endif
	  return 0;
	case 5:		/* B=C */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tB=C\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  copy_register(saturn.B, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.B);
#endif
	  return 0;
	case 6:		/* C=A */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tC=A\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  copy_register(saturn.C, saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.C);
#endif
	  return 0;
	case 7:		/* D=C */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tD=C\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  copy_register(saturn.D, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.D);
#endif
	  return 0;
	case 8:		/* B=A */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tB=A\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  copy_register(saturn.B, saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.B);
#endif
	  return 0;
	case 9:		/* C=B */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tC=B\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  copy_register(saturn.C, saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.C);
#endif
	  return 0;
	case 0xa:	/* A=C */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tA=C\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  copy_register(saturn.A, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.A);
#endif
	  return 0;
	case 0xb:	/* C=D */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tC=D\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  copy_register(saturn.C, saturn.D, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.D);
#endif
	  return 0;
	case 0xc:	/* ABEX */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tABEX\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  exchange_register(saturn.A, saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.A);
	  dump_register(saturn.B);
#endif
	  return 0;
	case 0xd:	/* BCEX */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tBCEX\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  exchange_register(saturn.B, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.B);
	  dump_register(saturn.C);
#endif
	  return 0;
	case 0xe:	/* ACEX */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tACEX\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  exchange_register(saturn.A, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.A);
	  dump_register(saturn.C);
#endif
	  return 0;
	case 0xf:	/* CDEX */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tCDEX\tA\n", saturn.PC);
#endif
	  saturn.PC += 2;
	  exchange_register(saturn.C, saturn.D, A_FIELD);
#ifdef DUMP_REGISTERS
	  dump_register(saturn.C);
	  dump_register(saturn.D);
#endif
	  return 0;
	default:
	  return 1;
      }
    case 0xe:
      switch (op2) {
        case 0:	/* A=A-B */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tA=A-B\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          sub_register(saturn.A, saturn.A, saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
#endif
          return 0;
        case 1:	/* B=B-C */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tB=B-C\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          sub_register(saturn.B, saturn.B, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.B);
#endif
          return 0;
        case 2:	/* C=C-A */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tC=C-A\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          sub_register(saturn.C, saturn.C, saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
#endif
          return 0;
        case 3:	/* D=D-C */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tD=D-C\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          sub_register(saturn.D, saturn.D, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.D);
#endif
          return 0;
        case 4:	/* A=A+1 */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tA=A+1\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
	  inc_register(saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
#endif
          return 0;
        case 5:	/* B=B+1 */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tB=B+1\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
	  inc_register(saturn.B, A_FIELD);
          return 0;
        case 6:	/* C=C+1 */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tC=C+1\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
	  inc_register(saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
#endif
          return 0;
        case 7:	/* D=D+1 */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tD=D+1\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
	  inc_register(saturn.D, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.D);
#endif
          return 0;
        case 8:	/* B=B-A */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tB=B-A\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          sub_register(saturn.B, saturn.B, saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.B);
#endif
          return 0;
        case 9:	/* C=C-B */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tC=C-B\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          sub_register(saturn.C, saturn.C, saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
#endif
          return 0;
        case 0xa:	/* A=A-C */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tA=A-C\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          sub_register(saturn.A, saturn.A, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
#endif
          return 0;
        case 0xb:	/* C=C-D */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tC=C-D\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          sub_register(saturn.C, saturn.C, saturn.D, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
#endif
          return 0;
        case 0xc:	/* A=B-A */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tA=B-A\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          sub_register(saturn.A, saturn.B, saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
#endif
          return 0;
        case 0xd:	/* B=C-B */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tB=C-B\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          sub_register(saturn.B, saturn.C, saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.B);
#endif
          return 0;
        case 0xe:	/* C=A-C */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tC=A-C\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          sub_register(saturn.C, saturn.A, saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
#endif
          return 0;
        case 0xf:	/* D=C-D */
#ifdef DEBUG_PRINT
	  dprintf("%.5lx\tD=C-D\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          sub_register(saturn.D, saturn.C, saturn.D, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.D);
#endif
          return 0;
        default:
          return 1;
      }
    case 0xf:
      switch (op2) {
        case 0:	/* ASL */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tASL\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          shift_left_register(saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
#endif
          return 0;
        case 1:	/* BSL */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tBSL\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          shift_left_register(saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.B);
#endif
          return 0;
        case 2:	/* CSL */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tCSL\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          shift_left_register(saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
#endif
          return 0;
        case 3:	/* DSL */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tDSL\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          shift_left_register(saturn.D, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.D);
#endif
          return 0;
        case 4:	/* ASR */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tASR\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          shift_right_register(saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
#endif
          return 0;
        case 5:	/* BSR */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tBSR\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          shift_right_register(saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.B);
#endif
          return 0;
        case 6:	/* CSR */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tCSR\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          shift_right_register(saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
#endif
          return 0;
        case 7:	/* DSR */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tDSR\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
          shift_right_register(saturn.D, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.D);
#endif
          return 0;
        case 8:	/* A=-A */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tA=-A\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
	  complement_2_register(saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
#endif
          return 0;
        case 9:	/* B=-B */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tB=-B\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
	  complement_2_register(saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.B);
#endif
          return 0;
        case 0xa:	/* C=-C */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tC=-C\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
	  complement_2_register(saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
#endif
          return 0;
        case 0xb:	/* D=-D */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tD=-D\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
	  complement_2_register(saturn.D, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.D);
#endif
          return 0;
        case 0xc:	/* A=-A-1 */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tA=-A-1\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
	  complement_1_register(saturn.A, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.A);
#endif
          return 0;
        case 0xd:	/* B=-B-1 */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tB=-B-1\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
	  complement_1_register(saturn.B, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.B);
#endif
          return 0;
        case 0xe:	/* C=-C-1 */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tC=-C-1\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
	  complement_1_register(saturn.C, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.C);
#endif
          return 0;
        case 0xf:	/* D=-D-1 */
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tD=-D-1\tA\n", saturn.PC);
#endif
          saturn.PC += 2;
	  complement_1_register(saturn.D, A_FIELD);
#ifdef DUMP_REGISTERS
          dump_register(saturn.D);
#endif
          return 0;
        default:
          return 1;
      }
    default:
      return 1;
  }
}

int
#ifdef __FunctionProto__
emulate(void)
#else
emulate()
#endif
{
  long op0, op1, op2, op3;
  int stop = 0;
  static unsigned long instructions = 0;
  long dsecs;
  struct timeval tv;

#ifdef SOLARIS
  gettimeofday(&tv);
#else
  gettimeofday(&tv, (struct timezone *)0);
#endif
  tv_t1.tv_sec = tv_t2.tv_sec = tv.tv_sec;
  tv_t1.tv_usec = tv_t2.tv_usec = tv.tv_usec;
  set_t1 = saturn.timer1;
  set_t2 = saturn.timer2;

  do {
    jumpaddr = 0;

    op0 = read_nibble(saturn.PC);

    switch (op0) {
      case 0:
        op1 = read_nibble(saturn.PC + 1);
	switch (op1) {
          case 0:	/* RTNSXM */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tRTNSXM\n", saturn.PC);
#endif
            saturn.XM = 1;
	    saturn.PC = pop_return_addr();
	    break;
          case 1:	/* RTN */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tRTN\n", saturn.PC);
#endif
	    saturn.PC = pop_return_addr();
	    break;
          case 2:	/* RTNSC */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tRTNSC\n", saturn.PC);
#endif
	    saturn.CARRY = 1;
	    saturn.PC = pop_return_addr();
	    break;
          case 3:	/* RTNCC */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tRTNCC\n", saturn.PC);
#endif
	    saturn.CARRY = 0;
	    saturn.PC = pop_return_addr();
	    break;
          case 4:	/* SETHEX */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tSETHEX\n", saturn.PC);
#endif
	    saturn.PC += 2;
	    saturn.hexmode = HEX;
	    break;
          case 5:	/* SETDEC */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tSETDEC\n", saturn.PC);
#endif
	    saturn.PC += 2;
	    saturn.hexmode = DEC;
	    break;
          case 6:	/* RSTK=C */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tRSTK=C\n", saturn.PC);
#endif
            jumpaddr = dat_to_addr(saturn.C);
     	    push_return_addr(jumpaddr);
            saturn.PC += 2;
	    break;
          case 7:	/* C=RSTK */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tC=RSTK\n", saturn.PC);
#endif
            saturn.PC += 2;
	    jumpaddr = pop_return_addr();
            addr_to_dat(jumpaddr, saturn.C);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.C);
#endif
	    break;
          case 8:	/* CLRST */
#ifdef DEBUG_PRINT
	    dprintf("%.5lx\tCLRST\n", saturn.PC);
#endif
            saturn.PC += 2;
            clear_status();
	    break;
          case 9:	/* C=ST */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tC=ST\n", saturn.PC);
#endif
            saturn.PC += 2;
            status_to_register(saturn.C);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.C);
#endif
	    break;
          case 0xa:	/* ST=C */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tST=C\n", saturn.PC);
#endif
            saturn.PC += 2;
            register_to_status(saturn.C);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.C);
#endif
	    break;
          case 0xb:	/* CSTEX */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tCSTEX\n", saturn.PC);
#endif
            saturn.PC += 2;
            swap_register_status(saturn.C);
#ifdef DUMP_REGISTERS
	    dump_register(saturn.C);
#endif
	    break;
          case 0xc:	/* P=P+1 */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tP=P+1\n", saturn.PC);
#endif
	    saturn.PC += 2;
            if (saturn.P == 0xf) {
              saturn.P = 0;
              saturn.CARRY = 1;
            } else {
              saturn.P += 1;
              saturn.CARRY = 0;
            }
#ifdef DUMP_REGISTERS
#ifdef DEBUG_PRINT
	    dprintf("P = %x", saturn.P);
#endif
#endif
	    break;
          case 0xd:	/* P=P-1 */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tP=P-1\n", saturn.PC);
#endif
	    saturn.PC += 2;
            if (saturn.P == 0) {
              saturn.P = 0xf;
              saturn.CARRY = 1;
            } else {
              saturn.P -= 1;
              saturn.CARRY = 0;
            }
#ifdef DUMP_REGISTERS
#ifdef DEBUG_PRINT
	    dprintf("P = %x", saturn.P);
#endif
#endif
	    break;
          case 0xe:
            op2 = read_nibble(saturn.PC + 2);
            op3 = read_nibble(saturn.PC + 3);
            switch(op3) {
              case 0:	/* A=A&B */
#ifdef DEBUG_PRINT
                dprintf("%.5lx\tA=A&B\t%s\n", saturn.PC, std_fields[op2]);
#endif
                saturn.PC += 4;
                and_register(saturn.A, saturn.A, saturn.B, op2);
#ifdef DUMP_REGISTERS
                dump_register(saturn.A);
#endif
                break;
              case 1:	/* B=B&C */
#ifdef DEBUG_PRINT
                dprintf("%.5lx\tB=B&C\t%s\n", saturn.PC, std_fields[op2]);
#endif
                saturn.PC += 4;
                and_register(saturn.B, saturn.B, saturn.C, op2);
#ifdef DUMP_REGISTERS
                dump_register(saturn.B);
#endif
                break;
              case 2:	/* C=C&A */
#ifdef DEBUG_PRINT
                dprintf("%.5lx\tC=C&A\t%s\n", saturn.PC, std_fields[op2]);
#endif
                saturn.PC += 4;
                and_register(saturn.C, saturn.C, saturn.A, op2);
#ifdef DUMP_REGISTERS
                dump_register(saturn.C);
#endif
                break;
              case 3:	/* D=D&C */
#ifdef DEBUG_PRINT
                dprintf("%.5lx\tD=D&C\t%s\n", saturn.PC, std_fields[op2]);
#endif
                saturn.PC += 4;
                and_register(saturn.D, saturn.D, saturn.C, op2);
#ifdef DUMP_REGISTERS
                dump_register(saturn.D);
#endif
                break;
              case 4:	/* B=B&A */
#ifdef DEBUG_PRINT
                dprintf("%.5lx\tB=B&A\t%s\n", saturn.PC, std_fields[op2]);
#endif
                saturn.PC += 4;
                and_register(saturn.B, saturn.B, saturn.A, op2);
#ifdef DUMP_REGISTERS
                dump_register(saturn.B);
#endif
                break;
              case 5:	/* C=C&B */
#ifdef DEBUG_PRINT
                dprintf("%.5lx\tC=C&B\t%s\n", saturn.PC, std_fields[op2]);
#endif
                saturn.PC += 4;
                and_register(saturn.C, saturn.C, saturn.B, op2);
#ifdef DUMP_REGISTERS
                dump_register(saturn.C);
#endif
                break;
              case 6:	/* A=A&C */
#ifdef DEBUG_PRINT
                dprintf("%.5lx\tA=A&C\t%s\n", saturn.PC, std_fields[op2]);
#endif
                saturn.PC += 4;
                and_register(saturn.A, saturn.A, saturn.C, op2);
#ifdef DUMP_REGISTERS
                dump_register(saturn.A);
#endif
                break;
              case 7:	/* C=C&D */
#ifdef DEBUG_PRINT
                dprintf("%.5lx\tC=C&D\t%s\n", saturn.PC, std_fields[op2]);
#endif
                saturn.PC += 4;
                and_register(saturn.C, saturn.C, saturn.D, op2);
#ifdef DUMP_REGISTERS
                dump_register(saturn.C);
#endif
                break;
              case 8:	/* A=A!B */
#ifdef DEBUG_PRINT
                dprintf("%.5lx\tA=A!B\t%s\n", saturn.PC, std_fields[op2]);
#endif
                saturn.PC += 4;
                or_register(saturn.A, saturn.A, saturn.B, op2);
#ifdef DUMP_REGISTERS
                dump_register(saturn.A);
#endif
                break;
              case 9:	/* B=B!C */
#ifdef DEBUG_PRINT
                dprintf("%.5lx\tB=B!C\t%s\n", saturn.PC, std_fields[op2]);
#endif
                saturn.PC += 4;
                or_register(saturn.B, saturn.B, saturn.C, op2);
#ifdef DUMP_REGISTERS
                dump_register(saturn.B);
#endif
                break;
              case 0xa:	/* C=C!A */
#ifdef DEBUG_PRINT
                dprintf("%.5lx\tC=C!A\t%s\n", saturn.PC, std_fields[op2]);
#endif
                saturn.PC += 4;
                or_register(saturn.C, saturn.C, saturn.A, op2);
#ifdef DUMP_REGISTERS
                dump_register(saturn.C);
#endif
                break;
              case 0xb:	/* D=D!C */
#ifdef DEBUG_PRINT
                dprintf("%.5lx\tD=D!C\t%s\n", saturn.PC, std_fields[op2]);
#endif
                saturn.PC += 4;
                or_register(saturn.D, saturn.D, saturn.C, op2);
#ifdef DUMP_REGISTERS
                dump_register(saturn.D);
#endif
                break;
              case 0xc:	/* B=B!A */
#ifdef DEBUG_PRINT
                dprintf("%.5lx\tB=B!A\t%s\n", saturn.PC, std_fields[op2]);
#endif
                saturn.PC += 4;
                or_register(saturn.B, saturn.B, saturn.A, op2);
#ifdef DUMP_REGISTERS
                dump_register(saturn.B);
#endif
                break;
              case 0xd:	/* C=C!B */
#ifdef DEBUG_PRINT
                dprintf("%.5lx\tC=C!B\t%s\n", saturn.PC, std_fields[op2]);
#endif
                saturn.PC += 4;
                or_register(saturn.C, saturn.C, saturn.B, op2);
#ifdef DUMP_REGISTERS
                dump_register(saturn.C);
#endif
                break;
              case 0xe:	/* A=A!C */
#ifdef DEBUG_PRINT
                dprintf("%.5lx\tA=A!C\t%s\n", saturn.PC, std_fields[op2]);
#endif
                saturn.PC += 4;
                or_register(saturn.A, saturn.A, saturn.C, op2);
#ifdef DUMP_REGISTERS
                dump_register(saturn.A);
#endif
                break;
              case 0xf:	/* C=C!D */
#ifdef DEBUG_PRINT
                dprintf("%.5lx\tC=C!D\t%s\n", saturn.PC, std_fields[op2]);
#endif
                saturn.PC += 4;
                or_register(saturn.C, saturn.C, saturn.D, op2);
#ifdef DUMP_REGISTERS
                dump_register(saturn.C);
#endif
                break;
              default:
                stop = 1;
                break;
            }
	    break;
          case 0xf:	/* RTI */
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tRTI\n", saturn.PC);
#endif
            do_return_interupt();
#ifdef DEBUG_INTERUPT
            fprintf(stderr, "saturn.PC = %.5lx: RTI\n", saturn.PC);
#endif
	    break;
          default:
            stop = 1;
            break;
        }
        break;
      case 1:
        stop = decode_group_1();
        break;
      case 2:
        op2 = read_nibble(saturn.PC + 1);
#ifdef DEBUG_PRINT
        dprintf("%.5lx\tP=\t%d\n", saturn.PC, op2);
#endif
        saturn.PC += 2;
        saturn.P = op2;
        break;
      case 3:
        op2 = read_nibble(saturn.PC + 1);
#ifdef DEBUG_PRINT
        dprintf("%.5lx\tLCHEX\t%s\n", saturn.PC,
                make_hexstr(saturn.PC + 2, op2 + 1));
#endif
        load_constant(saturn.C, op2 + 1, saturn.PC + 2);
        saturn.PC += 3 + op2;
#ifdef DUMP_REGISTERS
        dump_register(saturn.C);
#endif
        break;
      case 4:
        op2 = read_nibbles(saturn.PC + 1, 2);
        if (op2 == 0x02) {
#ifdef DEBUG_PRINT
          dprintf("%.5X\tNOP3\n", saturn.PC);
#endif
          saturn.PC += 3;
        } else if (saturn.CARRY != 0) {
          if (op2) {
            if (op2 & 0x80)
              op2 |= 0xffffff00;
            jumpaddr = (saturn.PC + op2 + 1) & 0xfffff;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tGOC\t%.5lx\n", saturn.PC, jumpaddr);
#endif
            saturn.PC = jumpaddr;
          } else {
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tRTNC\n", saturn.PC);
#endif
	    saturn.PC = pop_return_addr();
          }
        } else {
          saturn.PC += 3;
        }
        break;
      case 5:
        if (saturn.CARRY == 0) {
	  op2 = read_nibbles(saturn.PC + 1, 2);
          if (op2) {
            if (op2 & 0x80)
              op2 |= 0xffffff00;
            jumpaddr = (saturn.PC + op2 + 1) & 0xfffff;
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tGONC\t%.5lx\n", saturn.PC, jumpaddr);
#endif
            saturn.PC = jumpaddr;
          } else {
#ifdef DEBUG_PRINT
            dprintf("%.5lx\tRTNNC\n", saturn.PC);
#endif
	    saturn.PC = pop_return_addr();
          }
        } else {
          saturn.PC += 3;
        }
        break;
      case 6:
        op2 = read_nibbles(saturn.PC + 1, 3);
        if (op2 == 0x003) {
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tNOP4\n", saturn.PC);
#endif
          saturn.PC += 4;
        } else if(op2 == 0x004) {
          read_nibble(saturn.PC + 4);
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tNOP5\n", saturn.PC);
#endif
          saturn.PC += 5;
        } else {
          if (op2 & 0x800)
            op2 |= 0xfffff000;
          jumpaddr = (op2 + saturn.PC + 1) & 0xfffff;
#ifdef DEBUG_PRINT
          dprintf("%.5lx\tGOTO\t%.5lx\n", saturn.PC, jumpaddr);
#endif
          saturn.PC = jumpaddr;
        }
        break;
      case 7:
	op2 = read_nibbles(saturn.PC + 1, 3);
        if (op2 & 0x800)
          op2 |= 0xfffff000;
	jumpaddr = (op2 + saturn.PC + 4) & 0xfffff;
#ifdef DEBUG_PRINT
        dprintf("%.5lx\tGOSUB\t%.5lx\n", saturn.PC, jumpaddr);
#endif
	push_return_addr(saturn.PC+4);
        saturn.PC = jumpaddr;
        break;
      default:
        stop = decode_8_thru_f(op0);
        break;
    }

    check_devices();

    if (saturn.t1_instr == 0) {
#ifdef SOLARIS
      gettimeofday(&tv_t1);
#else
      gettimeofday(&tv_t1, (struct timezone *)0);
#endif
      set_t1 = saturn.timer1;
    }
    if (saturn.t2_instr == 0) {
#ifdef SOLARIS
      gettimeofday(&tv_t2);
#else
      gettimeofday(&tv_t2, (struct timezone *)0);
#endif
      set_t2 = saturn.timer2;
    }

    instructions++;

    if (got_alarm) {
      got_alarm = 0;
      GetEvent();
    }

    if ((instructions & 0x7f) == 0) {
      receive_char();
    }

    if ((++saturn.t2_instr & 0x1fff) == 0) {

#ifdef SOLARIS
      gettimeofday(&tv);
#else
      gettimeofday(&tv, (struct timezone *)0);
#endif
      dsecs = timerdiff(&tv, &tv_t2);
      if (saturn.t2_ctrl & 0x01) {
        saturn.timer2 = set_t2 - (dsecs * 128 / 15625);
      }
      if (dsecs > 200000) {
        tv_t2.tv_sec = tv.tv_sec;
        tv_t2.tv_usec = tv.tv_usec;
        set_t2 = saturn.timer2;
      }
      if ((saturn.t2_ctrl & 0x08) == 0 && saturn.timer2 <= 0) {
        if (saturn.t2_ctrl & 0x02) {
          saturn.t2_ctrl |= 0x08;
          do_interupt();
        }
      }

      dsecs = timerdiff(&tv, &tv_t1);
      saturn.timer1 = (set_t1 - (dsecs / 62500)) & 0x0f;
      if (dsecs > 200000) {
        tv_t1.tv_sec = tv.tv_sec;
        tv_t1.tv_usec = tv.tv_usec;
        set_t1 = saturn.timer1;
      }
      if ((saturn.t1_ctrl & 0x08) == 0 && saturn.timer1 == 0) {
        if (saturn.t1_ctrl & 0x02) {
          saturn.t1_ctrl |= 0x08;
          do_interupt();
        }
      }

    }

    if ((++saturn.t2_instr & 0xf) == 0) {
      if (saturn.t2_ctrl & 0x01) {
        saturn.timer2--;
      }
      if (saturn.timer2 == 0 && (saturn.t2_ctrl & 0x02)) {
        saturn.t2_ctrl |= 0x08;
        do_interupt();
      }
    }
    if ((++saturn.t1_instr & 0x1fff) == 0) {
      saturn.timer1 = (saturn.timer1 - 1) & 0xf;
      if (saturn.timer1 == 0 && (saturn.t1_ctrl & 0x02)) {
        saturn.t1_ctrl |= 0x08;
        do_interupt();
      }
    }

#ifdef DEBUG_SPEED
    if ((instructions & 0xfffff) == 0) {
      struct timeval tv;
      static long starttime = 0;
#ifdef SOLARIS
      gettimeofday(&tv);
#else
      gettimeofday(&tv, (struct timezone *)0);
#endif
      if (!starttime) {
        starttime = tv.tv_sec;
      }
      if (tv.tv_sec - starttime != 0) {
        fprintf(stdout, "\r%ld Instructions, %ld Instr/s        ",
                instructions, instructions / (tv.tv_sec - starttime));
        fflush(stdout);
      }
    }
#endif

  } while (stop == 0);

  if (stop) {
    fprintf(stderr, "PC = %.5lx, op = %x%x%x%x%x%x\n", saturn.PC,
            read_nibble(saturn.PC),
            read_nibble(saturn.PC + 1),
            read_nibble(saturn.PC + 2),
            read_nibble(saturn.PC + 3),
            read_nibble(saturn.PC + 4),
            read_nibble(saturn.PC + 5));
    exit(1);
  }

  return 0;
}
