/*
 *  This file is part of x48, an emulator of the HP-48sx Calculator.
 *  Copyright (C) 1994  Eddie C. Dost  (ecd@dressler.de)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Log: init.c,v $
 * Revision 1.4  1994/09/18  15:29:22  ecd
 * turned off unused rcsid message
 *
 * Revision 1.3  1994/09/13  16:57:00  ecd
 * changed to plain X11
 *
 * Revision 1.2  1994/08/31  18:23:21  ecd
 * changed display initialization.
 *
 * Revision 1.1  1994/08/26  11:09:02  ecd
 * Initial revision
 *
 * $Id: init.c,v 1.4 1994/09/18 15:29:22 ecd Exp ecd $
 */


#include "global.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>
#include <string.h>

#include "hp48.h"
#include "hp48_emu.h"
#include "device.h"
#include "debug.h"

#define ROWS		 70
#define COLUMNS		140

extern char lcd_buffer[ROWS][COLUMNS];

void
#ifdef __FunctionProto__
saturn_config_init(void)
#else
saturn_config_init()
#endif
{
  memset(&device, 0, sizeof(device));
  device.display_touched = 1;
  init_display();
  write_nibble(0x10f, 0xf);     /* CARDSTAT */
  write_nibble(0x111, 0x0);     /* RCS */
  write_nibble(0x112, 0x0);     /* TCS */
  write_nibble(0x11d, 0x0);     /* LBR */
}

int
#ifdef __FunctionProto__
read_rom(const char *fname)
#else
read_rom(fname)
const char *fname;
#endif
{
  FILE *in;

  if ((in = fopen(fname, "r")) == NULL) {
    return 0;
  }
  if ((saturn.rom = (unsigned char *)malloc(ROM_SIZE)) == NULL) {
    fclose(in);
    return 0;
  }
  if (fread(saturn.rom, 1, ROM_SIZE, in) != ROM_SIZE) {
    fclose(in);
    free(saturn.rom);
    return 0;
  }
  fclose(in);

  if ((saturn.ram = (unsigned char *)malloc(RAM_SIZE)) == NULL) {
    fprintf(stderr, "Can\'t malloc RAM\n");
    return 0;
  }
  memset(saturn.ram, 0, RAM_SIZE);

  if ((saturn.port1 = (unsigned char *)malloc(PORT1_SIZE)) == NULL) {
    fprintf(stderr, "Can\'t malloc PORT1\n");
    return 0;
  }
  memset(saturn.port1, 0, PORT1_SIZE);

  if ((saturn.port2 = (unsigned char *)malloc(PORT2_SIZE)) == NULL) {
    fprintf(stderr, "Can\'t malloc PORT2\n");
    return 0;
  }
  memset(saturn.port2, 0, PORT2_SIZE);

  return 1;
}

void
#ifdef __FunctionProto__
init_saturn(void)
#else
init_saturn()
#endif
{
  int i;

  debug_prints = 0;
  dump_registers = 0;
  saturn.PC = 0;
  saturn.P = 0;
  saturn.XM = saturn.SB = saturn.SR = saturn.MP = 0;
  for (i = 0; i < NR_PSTAT; i++) {
    saturn.PSTAT[i] = 0;
  }
  for (i = 0; i < 16; i++) {
    saturn.A[i] = 0;
    saturn.B[i] = 0;
    saturn.C[i] = 0;
    saturn.D[i] = 0;
    saturn.R0[i] = 0;
    saturn.R1[i] = 0;
    saturn.R2[i] = 0;
    saturn.R3[i] = 0;
    saturn.R4[i] = 0;
  }
  saturn.hexmode = HEX;
  saturn.rstkp = -1;
  for (i = 0; i < NR_RSTK; i++) {
    saturn.rstk[i] = 0;
  }
  saturn.D0 = 0;
  saturn.D1 = 0;
  saturn.CARRY = 0;
  for (i = 0; i < 3; i++) {
    saturn.IN[i] = 0;
    saturn.OUT[i] = 0;
  }
  saturn.IN[4] = 0;
  for (i = 0; i < NR_CONFIG; i++) {
    saturn.configs[i] = 0;
  }
  saturn.daisy_state = 0;
  saturn.ram32k = 0x70000;
  saturn.devices = 0x00100;
  saturn.intenable = 1;
  saturn.int_pending = 0;
  for (i = 0; i < 9; i++)
    saturn.keybuf.rows[i] = 0;
  dev_memory_init();
}

int
#ifdef __FunctionProto__
read_files(void)
#else
read_files()
#endif
{
  char *home;
  char path[256];
  char fnam[256];
  FILE *fp;

  if ((home = getenv("HOME")) == NULL) {
    fprintf(stderr, "can\'t getenv $HOME\n");
    return 0;
  }
  strcpy(path, home);
  strcat(path, "/.hp48/");

  strcpy(fnam, path);
  strcat(fnam, "hp48");
  if ((fp = fopen(fnam, "r")) == NULL) {
    fprintf(stderr, "can\'t open %s\n", fnam);
    return 0;
  }
  if (fread(&saturn, 1, sizeof(saturn), fp) != sizeof(saturn)) {
    fprintf(stderr, "can\'t read %s\n", fnam);
    fclose(fp);
    memset(&saturn, 0, sizeof(saturn));
    saturn.ram32k = 0x70000;
    saturn.devices = 0x00100;
    saturn.PC = 0x00000;
  }
  fclose(fp);

  saturn.rom = (unsigned char *)NULL;
  if ((saturn.rom = (unsigned char *)malloc(ROM_SIZE)) == NULL) {
    fprintf(stderr, "can\'t malloc rom[%d]\n", ROM_SIZE);
    exit (1);
  }
  strcpy(fnam, path);
  strcat(fnam, "rom");
  if ((fp = fopen(fnam, "r")) == NULL) {
    fprintf(stderr, "can\'t open %s\n", fnam);
    return 0;
  }
  if (fread(saturn.rom, 1, ROM_SIZE, fp) != ROM_SIZE) {
    fprintf(stderr, "can\'t read %s\n", fnam);
    fclose(fp);
    return 0;
  }
  fclose(fp);

  saturn_config_init();

  saturn.ram = (unsigned char *)NULL;
  if ((saturn.ram = (unsigned char *)malloc(RAM_SIZE)) == NULL) {
    fprintf(stderr, "can\'t malloc ram[%d]\n", RAM_SIZE);
    exit (1);
  }
  strcpy(fnam, path);
  strcat(fnam, "ram");
  if ((fp = fopen(fnam, "r")) == NULL) {
    fprintf(stderr, "can\'t open %s\n", fnam);
    return 0;
  }
  if (fread(saturn.ram, 1, RAM_SIZE, fp) != RAM_SIZE) {
    fprintf(stderr, "can\'t read %s\n", fnam);
    fclose(fp);
  }
  fclose(fp);

  saturn.port1 = (unsigned char *)NULL;
  if ((saturn.port1 = (unsigned char *)malloc(PORT1_SIZE)) == NULL) {
    fprintf(stderr, "can\'t malloc port1[%d]\n", PORT1_SIZE);
    exit (1);
  }
  strcpy(fnam, path);
  strcat(fnam, "port1");
  if ((fp = fopen(fnam, "r")) == NULL) {
    fprintf(stderr, "can\'t open %s\n", fnam);
    return 0;
  }
  if (fread(saturn.port1, 1, PORT1_SIZE, fp) != PORT1_SIZE) {
    fprintf(stderr, "can\'t read %s\n", fnam);
    fclose(fp);
  }
  fclose(fp);

  saturn.port2 = (unsigned char *)NULL;
  if ((saturn.port2 = (unsigned char *)malloc(PORT2_SIZE)) == NULL) {
    fprintf(stderr, "can\'t malloc port2[%d]\n", PORT2_SIZE);
    exit (1);
  }
  strcpy(fnam, path);
  strcat(fnam, "port2");
  if ((fp = fopen(fnam, "r")) == NULL) {
    fprintf(stderr, "can\'t open %s\n", fnam);
    return 0;
  }
  if (fread(saturn.port2, 1, PORT2_SIZE, fp) != PORT2_SIZE) {
    fprintf(stderr, "can\'t read %s\n", fnam);
    fclose(fp);
  }
  fclose(fp);

  return 1;
}

int
#ifdef __FunctionProto__
write_files(void)
#else
write_files()
#endif
{
  char *home;
  char path[256];
  char fnam[256];
  struct stat st;
  int make_dir;
  FILE *fp;

  make_dir = 0;
  if ((home = getenv("HOME")) != NULL) {
    strcpy(path, home);
    strcat(path, "/.hp48");
    if (stat(path, &st) == -1) {
      if (errno == ENOENT) {
        make_dir = 1;
      } else {
        fprintf(stderr, "can\'t stat %s, saving to /tmp\n", path);
        strcpy(path, "/tmp");
      }
    } else {
      if (!S_ISDIR(st.st_mode)) {
        fprintf(stderr, "%s is no directory, saving to /tmp\n", path);
        strcpy(path, "/tmp");
      }
    }
  } else {
    strcpy(path, "/tmp");
  }
  if (make_dir) {
    if (mkdir(path, 0777) == -1) {
      fprintf(stderr, "can\'t mkdir %s, saving to /tmp\n", path);
      strcpy(path, "/tmp");
    }
  }
  strcat(path, "/");
  strcpy(fnam, path);
  strcat(fnam, "hp48");
  if ((fp = fopen(fnam, "w")) == NULL) {
    fprintf(stderr, "can\'t open %s, no saving done\n", fnam);
    return 0;
  }
  if (fwrite(&saturn, 1, sizeof(saturn), fp) != sizeof(saturn)) {
    fprintf(stderr, "can\'t write to %s, no saving done\n", fnam);
    fclose(fp);
    return 0;
  }
  fclose(fp);

  strcpy(fnam, path);
  strcat(fnam, "rom");
  if ((fp = fopen(fnam, "w")) == NULL) {
    fprintf(stderr, "can\'t open %s, no saving done\n", fnam);
    return 0;
  }
  if (fwrite(saturn.rom, 1, ROM_SIZE, fp) != ROM_SIZE) {
    fprintf(stderr, "can\'t write to %s, no saving done\n", fnam);
    fclose(fp);
    return 0;
  }
  fclose(fp);

  strcpy(fnam, path);
  strcat(fnam, "ram");
  if ((fp = fopen(fnam, "w")) == NULL) {
    fprintf(stderr, "can\'t open %s, no saving done\n", fnam);
    return 0;
  }
  if (fwrite(saturn.ram, 1, RAM_SIZE, fp) != RAM_SIZE) {
    fprintf(stderr, "can\'t write to %s, no saving done\n", fnam);
    fclose(fp);
    return 0;
  }
  fclose(fp);

  strcpy(fnam, path);
  strcat(fnam, "port1");
  if ((fp = fopen(fnam, "w")) == NULL) {
    fprintf(stderr, "can\'t open %s, no saving done\n", fnam);
    return 0;
  }
  if (fwrite(saturn.port1, 1, PORT1_SIZE, fp) != PORT1_SIZE) {
    fprintf(stderr, "can\'t write to %s, no saving done\n", fnam);
    fclose(fp);
    return 0;
  }
  fclose(fp);

  strcpy(fnam, path);
  strcat(fnam, "port2");
  if ((fp = fopen(fnam, "w")) == NULL) {
    fprintf(stderr, "can\'t open %s, no saving done\n", fnam);
    return 0;
  }
  if (fwrite(saturn.port2, 1, PORT2_SIZE, fp) != PORT2_SIZE) {
    fprintf(stderr, "can\'t write to %s, no saving done\n", fnam);
    fclose(fp);
    return 0;
  }
  fclose(fp);

  return 1;
}

int 
#ifdef __FunctionProto__
init_emulator(void)
#else
init_emulator()
#endif
{
/*
  init_kbd();
*/
  serial_init();
  if (!read_files()) {
    if (!read_rom(CORE_FILE)) {
      fprintf(stderr, "%s: can\'t read corefile\n", progname);
      exit(1);
    }
    init_saturn();
  }
  init_display();
  
  return 0;
}

int
#ifdef __FunctionProto__
exit_emulator(void)
#else
exit_emulator()
#endif
{
  write_files();
  return 1;
}

